import re
import base64
from typing import Any, Dict
from datetime import date, datetime, time  # noqa: I251

from dlt.common.json import json
from dlt.common.pendulum import pendulum
from dlt.common.time import reduce_pendulum_datetime_precision

# use regex to escape characters in single pass
SQL_ESCAPE_DICT = {"'": "''", "\\": "\\\\", "\n": "\\n", "\r": "\\r"}


def _make_sql_escape_re(escape_dict: Dict[str, str]) -> re.Pattern:  # type: ignore[type-arg]
    return re.compile(
        "|".join([re.escape(k) for k in sorted(escape_dict, key=len, reverse=True)]),
        flags=re.DOTALL,
    )


SQL_ESCAPE_RE = _make_sql_escape_re(SQL_ESCAPE_DICT)


def _escape_extended(
    v: str,
    prefix: str = "E'",
    escape_dict: Dict[str, str] = None,
    escape_re: re.Pattern = None,  # type: ignore[type-arg]
) -> str:
    escape_dict = escape_dict or SQL_ESCAPE_DICT
    escape_re = escape_re or SQL_ESCAPE_RE
    return "{}{}{}".format(prefix, escape_re.sub(lambda x: escape_dict[x.group(0)], v), "'")


def escape_redshift_literal(v: Any) -> Any:
    if isinstance(v, str):
        # https://www.postgresql.org/docs/9.3/sql-syntax-lexical.html
        # looks like this is the only thing we need to escape for Postgres > 9.1
        # redshift keeps \ as escape character which is pre 9 behavior
        return _escape_extended(v, prefix="'")
    if isinstance(v, bytes):
        return f"from_hex('{v.hex()}')"
    if isinstance(v, (datetime, date, time)):
        return f"'{v.isoformat()}'"
    if isinstance(v, (list, dict)):
        return "json_parse(%s)" % _escape_extended(json.dumps(v), prefix="'")
    if v is None:
        return "NULL"

    return str(v)


def escape_postgres_literal(v: Any) -> Any:
    if isinstance(v, str):
        # we escape extended string which behave like the redshift string
        return _escape_extended(v)
    if isinstance(v, (datetime, date, time)):
        return f"'{v.isoformat()}'"
    if isinstance(v, (list, dict)):
        return _escape_extended(json.dumps(v))
    if isinstance(v, bytes):
        return f"'\\x{v.hex()}'"
    if v is None:
        return "NULL"

    return str(v)


def escape_duckdb_literal(v: Any) -> Any:
    if isinstance(v, str):
        # we escape extended string which behave like the redshift string
        return _escape_extended(v)
    if isinstance(v, (datetime, date, time)):
        return f"'{v.isoformat()}'"
    if isinstance(v, (list, dict)):
        return _escape_extended(json.dumps(v))
    if isinstance(v, bytes):
        return f"from_base64('{base64.b64encode(v).decode('ascii')}')"
    if v is None:
        return "NULL"

    return str(v)


def escape_lancedb_literal(v: Any) -> Any:
    if isinstance(v, str):
        # we escape extended string which behave like the redshift string
        return _escape_extended(v, prefix="'")
    if isinstance(v, (datetime, date, time)):
        return f"'{v.isoformat()}'"
    if isinstance(v, (list, dict)):
        return _escape_extended(json.dumps(v), prefix="'")
    # TODO: check how binaries are represented in fusion
    if isinstance(v, bytes):
        return f"from_base64('{base64.b64encode(v).decode('ascii')}')"
    if v is None:
        return "NULL"

    return str(v)


MS_SQL_ESCAPE_DICT = {
    "'": "''",
    "\n": "' + CHAR(10) + N'",
    "\r": "' + CHAR(13) + N'",
    "\t": "' + CHAR(9) + N'",
}
MS_SQL_ESCAPE_RE = _make_sql_escape_re(MS_SQL_ESCAPE_DICT)


def escape_mssql_literal(v: Any) -> Any:
    if isinstance(v, str):
        return _escape_extended(
            v, prefix="N'", escape_dict=MS_SQL_ESCAPE_DICT, escape_re=MS_SQL_ESCAPE_RE
        )
    if isinstance(v, (datetime, date, time)):
        return f"'{v.isoformat()}'"
    if isinstance(v, (list, dict)):
        return _escape_extended(
            json.dumps(v), prefix="N'", escape_dict=MS_SQL_ESCAPE_DICT, escape_re=MS_SQL_ESCAPE_RE
        )
    if isinstance(v, bytes):
        from dlt.destinations.impl.mssql.mssql import VARBINARY_MAX_N

        if len(v) <= VARBINARY_MAX_N:
            n = str(len(v))
        else:
            n = "MAX"
        return f"CONVERT(VARBINARY({n}), '{v.hex()}', 2)"

    if isinstance(v, bool):
        return str(int(v))
    if v is None:
        return "NULL"
    return str(v)


def escape_redshift_identifier(v: str) -> str:
    return '"' + v.replace('"', '""').replace("\\", "\\\\") + '"'


escape_postgres_identifier = escape_redshift_identifier
escape_athena_identifier = escape_postgres_identifier
escape_dremio_identifier = escape_postgres_identifier


def escape_hive_identifier(v: str) -> str:
    # https://cloud.google.com/bigquery/docs/reference/standard-sql/lexical
    return "`" + v.replace("\\", "\\\\").replace("`", "\\`") + "`"


def escape_snowflake_identifier(v: str) -> str:
    # Snowcase uppercase all identifiers unless quoted. Match this here so queries on information schema work without issue
    # See also https://docs.snowflake.com/en/sql-reference/identifiers-syntax#double-quoted-identifiers
    return escape_postgres_identifier(v)


def escape_snowflake_literal(v: Any) -> Any:
    """Escape string literals for Snowflake using standard SQL escaping.

    Snowflake uses '' to escape single quotes (not backslash escaping).
    """
    if isinstance(v, str):
        # Snowflake uses standard SQL escaping: ' -> ''
        return "'" + v.replace("'", "''") + "'"
    if isinstance(v, (datetime, date, time)):
        return f"'{v.isoformat()}'"
    if isinstance(v, (list, dict)):
        return "'" + json.dumps(v).replace("'", "''") + "'"
    if isinstance(v, bytes):
        return f"X'{v.hex()}'"
    return "NULL" if v is None else str(v)


escape_databricks_identifier = escape_hive_identifier


DATABRICKS_ESCAPE_DICT = {"'": "\\'", "\\": "\\\\", "\n": "\\n", "\r": "\\r"}


def escape_databricks_literal(v: Any) -> Any:
    if isinstance(v, str):
        return _escape_extended(v, prefix="'", escape_dict=DATABRICKS_ESCAPE_DICT)
    if isinstance(v, (datetime, date, time)):
        return f"'{v.isoformat()}'"
    if isinstance(v, (list, dict)):
        return _escape_extended(json.dumps(v), prefix="'", escape_dict=DATABRICKS_ESCAPE_DICT)
    if isinstance(v, bytes):
        return f"X'{v.hex()}'"
    return "NULL" if v is None else str(v)


# https://github.com/ClickHouse/ClickHouse/blob/master/docs/en/sql-reference/syntax.md#string
CLICKHOUSE_ESCAPE_DICT = {
    "'": "''",
    "\\": "\\\\",
    "\n": "\\n",
    "\t": "\\t",
    "\b": "\\b",
    "\f": "\\f",
    "\r": "\\r",
    "\0": "\\0",
    "\a": "\\a",
    "\v": "\\v",
}

CLICKHOUSE_ESCAPE_RE = _make_sql_escape_re(CLICKHOUSE_ESCAPE_DICT)


def escape_clickhouse_literal(v: Any) -> Any:
    if isinstance(v, str):
        return _escape_extended(
            v, prefix="'", escape_dict=CLICKHOUSE_ESCAPE_DICT, escape_re=CLICKHOUSE_ESCAPE_RE
        )
    if isinstance(v, (datetime, date, time)):
        return f"'{v.isoformat()}'"
    if isinstance(v, (list, dict)):
        return _escape_extended(
            json.dumps(v),
            prefix="'",
            escape_dict=CLICKHOUSE_ESCAPE_DICT,
            escape_re=CLICKHOUSE_ESCAPE_RE,
        )
    if isinstance(v, bytes):
        return f"'{v.hex()}'"
    return "NULL" if v is None else str(v)


def escape_clickhouse_identifier(v: str) -> str:
    return "`" + v.replace("`", "``").replace("\\", "\\\\") + "`"


# https://cloud.google.com/bigquery/docs/reference/standard-sql/lexical#string_and_bytes_literals
BIGQUERY_ESCAPE_DICT = {
    "'": "\\'",
    "\\": "\\\\",
    "\n": "\\n",
    "\r": "\\r",
    "\t": "\\t",
    "\b": "\\b",
    "\f": "\\f",
    "\a": "\\a",
    "\v": "\\v",
}

BIGQUERY_ESCAPE_RE = _make_sql_escape_re(BIGQUERY_ESCAPE_DICT)


def escape_bigquery_literal(v: Any) -> Any:
    if isinstance(v, str):
        return _escape_extended(
            v, prefix="'", escape_dict=BIGQUERY_ESCAPE_DICT, escape_re=BIGQUERY_ESCAPE_RE
        )
    if isinstance(v, (datetime, date, time)):
        return f"'{v.isoformat()}'"
    if isinstance(v, (list, dict)):
        return _escape_extended(
            json.dumps(v),
            prefix="'",
            escape_dict=BIGQUERY_ESCAPE_DICT,
            escape_re=BIGQUERY_ESCAPE_RE,
        )
    if isinstance(v, bytes):
        return f"FROM_BASE64('{base64.b64encode(v).decode('ascii')}')"
    if isinstance(v, bool):
        return str(v).upper()
    return "NULL" if v is None else str(v)


escape_bigquery_identifier = escape_hive_identifier


def format_datetime_literal(v: pendulum.DateTime, precision: int = 6, no_tz: bool = False) -> str:
    """Converts `v` to ISO string, optionally without timezone spec (in UTC) and with given `precision`"""
    if no_tz:
        v = v.in_timezone(tz="UTC").replace(tzinfo=None)
    v = reduce_pendulum_datetime_precision(v, precision)
    # yet another precision translation
    timespec: str = "microseconds"
    if precision < 6:
        timespec = "milliseconds"
    elif precision < 3:
        timespec = "seconds"
    return "'" + v.isoformat(sep=" ", timespec=timespec) + "'"


def format_bigquery_datetime_literal(
    v: pendulum.DateTime, precision: int = 6, no_tz: bool = False
) -> str:
    """Returns BigQuery-adjusted datetime literal by prefixing required `TIMESTAMP` indicator.

    Also works for Presto-based engines.
    """
    # https://cloud.google.com/bigquery/docs/reference/standard-sql/lexical#timestamp_literals
    return "TIMESTAMP " + format_datetime_literal(v, precision, no_tz)


def format_clickhouse_datetime_literal(
    v: pendulum.DateTime, precision: int = 6, no_tz: bool = False
) -> str:
    """Returns clickhouse compatibel function"""
    datetime = format_datetime_literal(v, precision, True)
    return f"toDateTime64({datetime}, {precision}, '{v.tzinfo}')"



# /// script
# [tool.marimo.display]
# theme = "light"
# ///
# flake8: noqa: F841
# mypy: disable-error-code=no-untyped-def

import marimo

__generated_with = "0.13.9"
app = marimo.App(
    width="medium", app_title="dlt workspace dashboard", css_file="dlt_dashboard_styles.css"
)

with app.setup:
    from typing import Any, Dict, List, cast

    from dlt._workspace.helpers.dashboard.typing import (
        TLoadItem,
        TPipelineListItem,
        TQueryHistoryItem,
        TTableListItem,
    )

    import marimo as mo

    import dlt
    import pyarrow
    from dlt.common import logger
    from dlt._workspace.helpers.dashboard import strings
    from dlt._workspace.helpers.dashboard import utils

    from dlt._workspace.helpers.dashboard.config import DashboardConfiguration
    from dlt.common.configuration.specs.pluggable_run_context import ProfilesRunContext
    from dlt._workspace.run_context import switch_profile


@app.cell(hide_code=True)
def home(
    dlt_profile_select: mo.ui.dropdown,
    dlt_all_pipelines: List[TPipelineListItem],
    dlt_pipeline_select: mo.ui.multiselect,
    dlt_pipelines_dir: str,
    dlt_refresh_button: mo.ui.run_button,
    dlt_pipeline_name: str,
    dlt_file_watcher: Any,  # marimo internal watcher type
):
    """Displays the welcome page or pipeline dashboard depending on selection."""

    # NOTE: keep these two lines for refreshing
    dlt_refresh_button
    dlt_file_watcher

    # returned by cell
    dlt_pipeline: dlt.Pipeline = None
    dlt_config: DashboardConfiguration = None
    _result: List[mo.Html] = []

    if dlt_pipeline_name:
        try:
            dlt_pipeline = utils.pipeline.get_pipeline(dlt_pipeline_name, dlt_pipelines_dir)
            dlt_config = utils.pipeline.resolve_dashboard_config(dlt_pipeline)
        except Exception:
            _result = utils.home.render_pipeline_header_row(
                dlt_pipeline_name, dlt_profile_select, dlt_pipeline_select, [dlt_refresh_button]
            )
            _result.append(
                utils.ui.error_callout(
                    strings.home_error_attach_pipeline.format(dlt_pipeline_name),
                )
            )

        if dlt_pipeline:
            try:
                _result = utils.home.render_pipeline_home(
                    dlt_profile_select,
                    dlt_pipeline,
                    dlt_pipeline_select,
                    dlt_pipelines_dir,
                    dlt_refresh_button,
                    dlt_pipeline_name,
                )
            except Exception:
                _result = [
                    utils.ui.error_callout(
                        strings.home_error_rendering_pipeline,
                        strings.home_error_rendering_pipeline_detail,
                    )
                ]
    else:
        try:
            dlt_config = utils.pipeline.resolve_dashboard_config(None)
            _result = utils.home.render_workspace_home(
                dlt_profile_select,
                dlt_all_pipelines,
                dlt_pipeline_select,
                dlt_pipelines_dir,
                dlt_config,
            )
        except Exception:
            _result = [
                utils.ui.error_callout(
                    strings.home_error_rendering_home,
                )
            ]
    mo.vstack(_result) if _result else None
    return (dlt_pipeline, dlt_config)


@app.cell(hide_code=True)
def section_info(
    dlt_pipeline: dlt.Pipeline,
    dlt_section_info_switch: mo.ui.switch,
    dlt_all_pipelines: List[TPipelineListItem],
    dlt_config: DashboardConfiguration,
    dlt_pipelines_dir: str,
):
    """
    Overview page of currently selected pipeline
    """

    _result, _show = utils.ui.section(strings.overview, dlt_pipeline, dlt_section_info_switch)

    if _show:
        _result += [
            utils.ui.dlt_table(
                utils.pipeline.pipeline_details(dlt_config, dlt_pipeline, dlt_pipelines_dir)
            ),
        ]
        _result.append(
            utils.ui.title_and_subtitle(
                strings.overview_remote_state_title, strings.overview_remote_state_subtitle
            )
        )
        _result.append(
            mo.accordion(
                {
                    strings.overview_remote_state_button: utils.ui.dlt_table(
                        utils.pipeline.remote_state_details(dlt_pipeline),
                    )
                },
                lazy=True,
            )
        )
    mo.vstack(_result) if _result else None
    return


@app.cell(hide_code=True)
def section_schema(
    dlt_config: DashboardConfiguration,
    dlt_pipeline: dlt.Pipeline,
    dlt_schema_show_child_tables: mo.ui.switch,
    dlt_schema_show_custom_hints: mo.ui.switch,
    dlt_schema_show_dlt_columns: mo.ui.switch,
    dlt_schema_show_dlt_tables: mo.ui.switch,
    dlt_schema_show_other_hints: mo.ui.switch,
    dlt_schema_show_type_hints: mo.ui.switch,
    dlt_schema_table_list: mo.ui.table,
    dlt_section_schema_switch: mo.ui.switch,
    dlt_schema_select: mo.ui.multiselect,
    dlt_selected_schema_name: str,
):
    """
    Show schema of the currently selected pipeline
    """

    _result, _show = utils.ui.section(strings.schema, dlt_pipeline, dlt_section_schema_switch)

    if _show and dlt_schema_table_list is None:
        _result.append(
            mo.callout(
                mo.md(strings.schema_no_default_available_text),
                kind="warn",
            )
        )
    elif _show:
        # build table overview
        _result.append(
            mo.hstack(
                [dlt_schema_select, dlt_schema_show_dlt_tables, dlt_schema_show_child_tables],
                justify="start",
            )
        )
        _result.append(dlt_schema_table_list)

        # add table details
        _result.append(utils.ui.title_and_subtitle(strings.schema_table_details_title))
        _result.append(
            mo.hstack(
                [
                    dlt_schema_show_dlt_columns,
                    dlt_schema_show_type_hints,
                    dlt_schema_show_other_hints,
                    dlt_schema_show_custom_hints,
                ],
                justify="start",
            ),
        )

        for _table in cast(List[TTableListItem], dlt_schema_table_list.value):
            _table_name = _table["name"]
            _result.append(mo.md(strings.schema_table_columns_title.format(_table_name)))
            _columns_list = utils.schema.create_column_list(
                dlt_config,
                dlt_pipeline,
                _table_name,
                dlt_selected_schema_name,
                show_internals=dlt_schema_show_dlt_columns.value,
                show_type_hints=dlt_schema_show_type_hints.value,
                show_other_hints=dlt_schema_show_other_hints.value,
                show_custom_hints=dlt_schema_show_custom_hints.value,
            )
            _result.append(utils.ui.dlt_table(_columns_list))

        # add raw schema
        _result.append(utils.ui.title_and_subtitle(strings.schema_raw_yaml_title))
        _result.append(
            mo.accordion(
                {
                    strings.schema_show_raw_yaml_text: mo.ui.code_editor(
                        dlt_pipeline.schemas[dlt_selected_schema_name].to_pretty_yaml(),
                        language="yaml",
                    )
                }
            )
        )
    mo.vstack(_result) if _result else None
    return


@app.cell(hide_code=True)
def ui_data_quality_controls(
    dlt_pipeline: dlt.Pipeline,
    dlt_section_data_quality_switch: mo.ui.switch,
):
    """Create data quality filter controls (separate cell for marimo reactivity)."""
    dlt_data_quality_show_failed_filter: mo.ui.checkbox = None
    dlt_data_quality_table_filter: mo.ui.dropdown = None
    dlt_data_quality_rate_filter: mo.ui.slider = None
    dlt_data_quality_checks_arrow = None

    if utils.home.detect_dlt_hub() and dlt_pipeline:
        (
            dlt_data_quality_show_failed_filter,
            dlt_data_quality_table_filter,
            dlt_data_quality_rate_filter,
            dlt_data_quality_checks_arrow,
        ) = utils.data_quality.create_dq_controls(dlt_pipeline)

    return (
        dlt_data_quality_show_failed_filter,
        dlt_data_quality_table_filter,
        dlt_data_quality_rate_filter,
        dlt_data_quality_checks_arrow,
    )


@app.cell(hide_code=True)
def section_browse_data_table_list(
    dlt_clear_result_cache: mo.ui.run_button,
    dlt_data_table_list: mo.ui.table,
    dlt_pipeline: dlt.Pipeline,
    dlt_restrict_to_last_1000: mo.ui.switch,
    dlt_schema_show_child_tables: mo.ui.switch,
    dlt_schema_show_dlt_tables: mo.ui.switch,
    dlt_schema_show_row_counts: mo.ui.run_button,
    dlt_section_browse_data_switch: mo.ui.switch,
    dlt_schema_select: mo.ui.multiselect,
    dlt_selected_schema_name: str,
):
    """
    Show data of the currently selected pipeline
    """

    _result, _show = utils.ui.section(
        strings.browse_data, dlt_pipeline, dlt_section_browse_data_switch
    )

    dlt_query_editor: mo.ui.code_editor = None
    dlt_run_query_button: mo.ui.run_button = None
    if _show and dlt_data_table_list is not None:
        _result.append(
            mo.hstack(
                [
                    dlt_schema_select,
                    dlt_schema_show_row_counts,
                    dlt_schema_show_dlt_tables,
                    dlt_schema_show_child_tables,
                ],
                justify="start",
            ),
        )
        _result.append(dlt_data_table_list)

        _sql_query = ""
        if dlt_data_table_list.value:
            _selected_table = cast(List[TTableListItem], dlt_data_table_list.value)[0]
            _table_name = _selected_table["name"]

            if _state_widget := utils.schema.build_resource_state_widget(
                dlt_pipeline, dlt_selected_schema_name, _table_name
            ):
                _result.append(_state_widget)

            _sql_query, _error_message, _traceback_string = (
                utils.queries.get_default_query_for_table(
                    dlt_pipeline,
                    dlt_selected_schema_name,
                    _table_name,
                    dlt_restrict_to_last_1000.value,
                )
            )
        _placeholder, _error_message, _traceback_string = (
            utils.queries.get_example_query_for_dataset(
                dlt_pipeline,
                dlt_selected_schema_name,
            )
        )

        if _error_message:
            _result.append(
                utils.ui.error_callout(
                    strings.browse_data_error_text + _error_message,
                    traceback_string=_traceback_string,
                )
            )
        else:
            dlt_query_editor = mo.ui.code_editor(
                language="sql",
                placeholder=_placeholder,
                value=_sql_query,
                debounce=True,
            )

            dlt_run_query_button = mo.ui.run_button(
                label=utils.ui.small(strings.browse_data_run_query_button),
                tooltip=strings.browse_data_run_query_tooltip,
            )

            _result += [
                mo.md(utils.ui.small(strings.browse_data_explorer_title)),
                mo.hstack(
                    [dlt_restrict_to_last_1000],
                    justify="start",
                ),
                dlt_query_editor,
            ]

            _result.append(
                mo.hstack([dlt_run_query_button, dlt_clear_result_cache], justify="start")
            )
    elif _show:
        # here we also use the no schemas text, as it is appropriate for the case where we have no table information.
        _result.append(utils.ui.error_callout(strings.schema_no_default_available_text))
    mo.vstack(_result) if _result else None
    return dlt_query_editor, dlt_run_query_button


@app.cell(hide_code=True)
def section_browse_data_query_result(
    dlt_data_table_list: mo.ui.table,
    dlt_pipeline: dlt.Pipeline,
    dlt_query_editor: mo.ui.code_editor,
    dlt_run_query_button: mo.ui.run_button,
    dlt_section_browse_data_switch: mo.ui.switch,
    dlt_clear_result_cache: mo.ui.run_button,
    dlt_get_last_query_result,
    dlt_set_last_query_result,
    dlt_set_query_history,
    dlt_get_query_history,
):
    """
    Execute the query in the editor
    """

    _result = []

    dlt_query_history_table: mo.ui.table = None
    dlt_query: str = None

    if (
        dlt_pipeline
        and dlt_section_browse_data_switch.value
        and dlt_data_table_list is not None
        and dlt_query_editor is not None
    ):
        _result.append(utils.ui.title_and_subtitle(strings.browse_data_query_result_title))
        _error_message: str = None
        with mo.status.spinner(title=strings.browse_data_loading_spinner_text):
            if dlt_query_editor.value and (dlt_run_query_button.value):
                if dlt_clear_result_cache.value:
                    utils.queries.clear_query_cache()
                dlt_query = dlt_query_editor.value
                _query_result, _error_message, _traceback_string = utils.queries.get_query_result(
                    dlt_pipeline, dlt_query
                )
                dlt_set_last_query_result(_query_result)

            # display error message if encountered
            if _error_message:
                _result.append(
                    utils.ui.error_callout(
                        strings.browse_data_query_error + _error_message,
                        traceback_string=_traceback_string,
                    )
                )

        # always display result table
        _last_result = dlt_get_last_query_result()
        if _last_result is not None:
            _result.append(utils.ui.dlt_table(_last_result, freeze_column=None))

        # update query history if there was no error
        if _last_result is not None and not _error_message:
            if dlt_query:
                dlt_set_query_history(
                    utils.queries.update_query_history(
                        dlt_get_query_history(), dlt_query, _last_result.shape[0]
                    )
                )

    # provide query history table
    if _query_history := dlt_get_query_history():
        dlt_query_history_table = utils.ui.dlt_table(
            [{"query": q, "row_count": _query_history[q]} for q in _query_history],
            selection="multi",
            freeze_column=None,
        )
    mo.vstack(_result) if _result else None
    return dlt_query_history_table


@app.cell(hide_code=True)
def section_browse_data_query_history(
    dlt_pipeline: dlt.Pipeline,
    dlt_query_history_table: mo.ui.table,
    dlt_section_browse_data_switch: mo.ui.switch,
):
    """
    Show the query history
    """

    _result: List[mo.Html] = []
    if (
        dlt_pipeline
        and dlt_section_browse_data_switch.value
        and dlt_query_history_table is not None
    ):
        _result.append(
            utils.ui.title_and_subtitle(
                strings.browse_data_query_history_title, strings.browse_data_query_history_subtitle
            )
        )
        _result.append(dlt_query_history_table)

        for _r in cast(List[TQueryHistoryItem], dlt_query_history_table.value):
            _query = _r["query"]
            _q_result, _q_error, _q_traceback = utils.queries.get_query_result(dlt_pipeline, _query)
            _result.append(mo.md(utils.ui.small(f"```{_query}```")))
            if _q_error:
                _result.append(utils.ui.error_callout(_q_error, traceback_string=_q_traceback))
            else:
                _result.append(utils.ui.dlt_table(_q_result, freeze_column=None))
    mo.vstack(_result) if _result else None
    return


@app.cell(hide_code=True)
def section_data_quality(
    dlt_pipeline: dlt.Pipeline,
    dlt_section_data_quality_switch: mo.ui.switch,
    dlt_data_quality_show_failed_filter: mo.ui.checkbox,
    dlt_data_quality_table_filter: mo.ui.dropdown,
    dlt_data_quality_rate_filter: mo.ui.slider,
    dlt_data_quality_checks_arrow,
):
    """Show data quality of the currently selected pipeline (only if dlt.hub is installed)."""
    dlt_data_quality_show_raw_table_switch = None
    if not utils.home.detect_dlt_hub():
        _result: List[mo.Html] = []
    else:
        _result, _show = utils.ui.section(
            strings.data_quality, dlt_pipeline, dlt_section_data_quality_switch
        )
        if _show:
            _dq_widgets, dlt_data_quality_show_raw_table_switch = (
                utils.data_quality.build_dq_section(
                    dlt_pipeline,
                    dlt_data_quality_show_failed_filter,
                    dlt_data_quality_table_filter,
                    dlt_data_quality_rate_filter,
                    dlt_data_quality_checks_arrow,
                )
            )
            _result.extend(_dq_widgets)
        else:
            dlt_data_quality_show_raw_table_switch = None
    mo.vstack(_result) if _result else None
    return dlt_data_quality_show_raw_table_switch


@app.cell(hide_code=True)
def section_data_quality_raw_table(
    dlt_pipeline: dlt.Pipeline,
    dlt_section_data_quality_switch: mo.ui.switch,
    dlt_data_quality_show_raw_table_switch: mo.ui.switch,
    dlt_get_last_query_result,
    dlt_set_last_query_result,
):
    """Display the raw data quality checks table with _dlt_load_id column."""
    _result = []

    if (
        dlt_pipeline
        and dlt_section_data_quality_switch.value
        and dlt_data_quality_show_raw_table_switch is not None
        and dlt_data_quality_show_raw_table_switch.value
    ):
        _result = utils.data_quality.build_dq_raw_table(
            dlt_pipeline, dlt_get_last_query_result, dlt_set_last_query_result
        )
    mo.vstack(_result) if _result else None
    return


@app.cell(hide_code=True)
def section_state(
    dlt_pipeline: dlt.Pipeline,
    dlt_section_state_switch: mo.ui.switch,
):
    """
    Show state of the currently selected pipeline
    """
    _result, _show = utils.ui.section(strings.state, dlt_pipeline, dlt_section_state_switch)

    if _show:
        _result.append(
            mo.json(
                dlt_pipeline.state,  # type: ignore[arg-type]
            ),
        )
    mo.vstack(_result) if _result else None
    return


@app.cell(hide_code=True)
def section_trace(
    dlt_pipeline: dlt.Pipeline,
    dlt_section_trace_switch: mo.ui.switch,
    dlt_trace_steps_table: mo.ui.table,
    dlt_config: DashboardConfiguration,
):
    """
    Show last trace of the currently selected pipeline
    """

    _result, _show = utils.ui.section(strings.trace, dlt_pipeline, dlt_section_trace_switch)

    if _show:
        try:
            _result.extend(
                utils.trace.build_trace_section(dlt_config, dlt_pipeline, dlt_trace_steps_table)
            )
        except Exception as exc:
            _result.append(
                utils.ui.error_callout(
                    strings.trace_error_building_section.format(exc),
                )
            )
    mo.vstack(_result) if _result else None
    return


@app.cell(hide_code=True)
def section_loads(
    dlt_config: DashboardConfiguration,
    dlt_clear_result_cache: mo.ui.run_button,
    dlt_pipeline: dlt.Pipeline,
    dlt_restrict_to_last_1000: mo.ui.switch,
    dlt_section_loads_switch: mo.ui.switch,
):
    """
    Show loads of the currently selected pipeline
    """

    _result, _show = utils.ui.section(strings.loads, dlt_pipeline, dlt_section_loads_switch)

    if _show:
        _result.append(mo.hstack([dlt_restrict_to_last_1000], justify="start"))
        with mo.status.spinner(title=strings.loads_loading_spinner_text):
            _loads_data, _error_message, _traceback_string = utils.queries.get_loads(
                dlt_config,
                dlt_pipeline,
                limit=1000 if dlt_restrict_to_last_1000.value else None,
            )
            dlt_loads_table = utils.ui.dlt_table(_loads_data, selection="single")
            if _error_message:
                _result.append(
                    utils.ui.error_callout(
                        strings.loads_loading_failed_text + _error_message,
                        traceback_string=_traceback_string,
                    )
                )
            _result.append(dlt_loads_table)
            _result.append(dlt_clear_result_cache)
    mo.vstack(_result) if _result else None
    return (dlt_loads_table,)


@app.cell(hide_code=True)
def section_loads_results(
    dlt_loads_table: mo.ui.table,
    dlt_pipeline: dlt.Pipeline,
    dlt_section_loads_switch: mo.ui.switch,
):
    """
    Show details of the currently selected load
    """
    _result = []

    if (
        dlt_pipeline
        and dlt_section_loads_switch.value
        and dlt_loads_table is not None
        and dlt_loads_table.value
    ):
        _selected_load = cast(List[TLoadItem], dlt_loads_table.value)[0]
        try:
            _result.extend(utils.queries.build_load_details(dlt_pipeline, _selected_load))
        except Exception:
            _result.append(utils.ui.error_callout(strings.loads_details_error_text))
    mo.vstack(_result) if _result else None
    return


@app.cell(hide_code=True)
def section_ibis_backend(
    dlt_pipeline: dlt.Pipeline,
    dlt_section_ibis_browser_switch: mo.ui.switch,
):
    """
    Connects to ibis backend and makes it available in the datasources panel.
    Only shown in edit mode since it is a developer tool.
    """
    mo.stop(mo.app_meta().mode != "edit")

    _result, _show = utils.ui.section(
        strings.ibis_backend, dlt_pipeline, dlt_section_ibis_browser_switch
    )

    if _show:
        try:
            with mo.status.spinner(title=strings.ibis_backend_connecting_spinner_text):
                con = dlt_pipeline.dataset().ibis(read_only=True)
            _result.append(
                mo.callout(mo.vstack([mo.md(strings.ibis_backend_connected_text)]), kind="success")
            )
        except Exception as exc:
            _result.append(utils.ui.error_callout(strings.ibis_backend_error_text + str(exc)))
    mo.vstack(_result) if _result else None
    return


@app.cell(hide_code=True)
def utils_discover_pipelines(
    dlt_profile_select: mo.ui.dropdown,
    mo_cli_arg_pipelines_dir: str,
    mo_cli_arg_pipeline: str,
    mo_query_var_pipeline_name: str,
):
    """
    Discovers local pipelines and returns a multiselect widget to select one of the pipelines
    """
    from dlt._workspace.cli.utils import list_local_pipelines

    # sync from runtime if enabled
    _tmp_config = utils.pipeline.resolve_dashboard_config(None)
    if _tmp_config.sync_from_runtime:
        from dlt._workspace.helpers.runtime.runtime_artifacts import sync_from_runtime

        with mo.status.spinner(title=strings.home_syncing_spinner_text):
            sync_from_runtime()

    _run_context = dlt.current.run_context()
    if (
        isinstance(_run_context, ProfilesRunContext)
        and not _run_context.profile == dlt_profile_select.value
    ):
        switch_profile(dlt_profile_select.value)

    # discover pipelines and build selector
    dlt_pipelines_dir: str = ""
    dlt_all_pipelines: List[TPipelineListItem] = []
    dlt_pipelines_dir, dlt_all_pipelines = list_local_pipelines(
        mo_cli_arg_pipelines_dir,
        additional_pipelines=[mo_cli_arg_pipeline, mo_query_var_pipeline_name],
    )

    dlt_pipeline_select: mo.ui.multiselect = mo.ui.multiselect(
        options=[p["name"] for p in dlt_all_pipelines],
        value=(
            [mo_query_var_pipeline_name]
            if mo_query_var_pipeline_name
            else ([mo_cli_arg_pipeline] if mo_cli_arg_pipeline else None)
        ),
        max_selections=1,
        label=strings.app_pipeline_select_label,
        on_change=lambda value: mo.query_params().set("pipeline", str(value[0]) if value else None),
    )

    return dlt_all_pipelines, dlt_pipeline_select, dlt_pipelines_dir


@app.cell(hide_code=True)
def utils_discover_profiles(mo_query_var_profile: str, mo_cli_arg_profile: str):
    """Discover profiles and return a single-select multiselect, similar to pipelines."""
    run_context = dlt.current.run_context()

    # Default (non-profile-aware) output
    dlt_profile_select = mo.ui.dropdown(
        options=[], value=None, label=strings.app_profile_select_label
    )
    selected_profile = None

    if isinstance(run_context, ProfilesRunContext):
        options = run_context.configured_profiles() or []
        current = run_context.profile if options and run_context.profile in options else None

        selected_profile = current
        if mo_query_var_profile and mo_query_var_profile in options:
            selected_profile = mo_query_var_profile
        elif mo_cli_arg_profile and mo_cli_arg_profile in options:
            selected_profile = mo_cli_arg_profile

        def _on_profile_change(v: str) -> None:
            mo.query_params().set("profile", v)

        dlt_profile_select = mo.ui.dropdown(
            options=options,
            value=selected_profile,
            label=strings.app_profile_select_label,
            on_change=_on_profile_change,
            searchable=True,
        )

    return dlt_profile_select, selected_profile


@app.cell(hide_code=True)
def utils_discover_schemas(dlt_pipeline: dlt.Pipeline):
    """
    Create schema multiselect widget
    """
    schemas = dlt_pipeline.schemas.values() if (dlt_pipeline and dlt_pipeline.schemas) else []
    dlt_schema_select: mo.ui.dropdown = mo.ui.dropdown(
        options=[s.name for s in schemas],
        value=(
            dlt_pipeline.default_schema_name
            if (dlt_pipeline and dlt_pipeline.default_schema_name)
            else None
        ),
        label=strings.app_schema_select_label,
    )
    return dlt_schema_select


@app.cell(hide_code=True)
def state_and_cache_reset(
    dlt_clear_result_cache: mo.ui.run_button,
    dlt_pipeline: dlt.Pipeline,
):
    """State variables for query results and history, plus cache clearing."""

    dlt_get_last_query_result, dlt_set_last_query_result = mo.state(pyarrow.table({}))
    # query history: maps query text to row count for the history sidebar
    dlt_get_query_history, dlt_set_query_history = mo.state(cast(Dict[str, int], {}))

    if dlt_clear_result_cache.value:
        utils.queries.clear_query_cache()

    return


@app.cell(hide_code=True)
def ui_controls(mo_cli_arg_with_test_identifiers: bool):
    """
    Control elements for various parts of the app
    """

    dlt_refresh_button: mo.ui.run_button = mo.ui.run_button(
        label=utils.ui.small(strings.app_refresh_button),
    )

    # page switches
    dlt_section_sync_switch: mo.ui.switch = mo.ui.switch(
        value=True, label="sync" if mo_cli_arg_with_test_identifiers else ""
    )
    dlt_section_info_switch: mo.ui.switch = mo.ui.switch(
        value=False, label="overview" if mo_cli_arg_with_test_identifiers else ""
    )
    dlt_section_schema_switch: mo.ui.switch = mo.ui.switch(
        value=False, label="schema" if mo_cli_arg_with_test_identifiers else ""
    )
    dlt_section_browse_data_switch: mo.ui.switch = mo.ui.switch(
        value=False, label="data" if mo_cli_arg_with_test_identifiers else ""
    )
    dlt_section_state_switch: mo.ui.switch = mo.ui.switch(
        value=False, label="state" if mo_cli_arg_with_test_identifiers else ""
    )
    dlt_section_trace_switch: mo.ui.switch = mo.ui.switch(
        value=False, label="trace" if mo_cli_arg_with_test_identifiers else ""
    )
    dlt_section_loads_switch: mo.ui.switch = mo.ui.switch(
        value=False, label="loads" if mo_cli_arg_with_test_identifiers else ""
    )
    dlt_section_ibis_browser_switch: mo.ui.switch = mo.ui.switch(
        value=False, label="ibis" if mo_cli_arg_with_test_identifiers else ""
    )
    dlt_section_data_quality_switch: mo.ui.switch = mo.ui.switch(
        value=False, label="data_quality" if mo_cli_arg_with_test_identifiers else ""
    )

    # other switches
    dlt_schema_show_dlt_tables: mo.ui.switch = mo.ui.switch(
        label=utils.ui.small(strings.ui_show_dlt_tables)
    )
    dlt_schema_show_child_tables: mo.ui.switch = mo.ui.switch(
        label=utils.ui.small(strings.ui_show_child_tables), value=True
    )
    dlt_schema_show_row_counts: mo.ui.run_button = mo.ui.run_button(
        label=utils.ui.small(strings.ui_load_row_counts)
    )
    dlt_schema_show_dlt_columns: mo.ui.switch = mo.ui.switch(
        label=utils.ui.small(strings.ui_show_dlt_columns)
    )
    dlt_schema_show_type_hints: mo.ui.switch = mo.ui.switch(
        label=utils.ui.small(strings.ui_show_type_hints), value=True
    )
    dlt_schema_show_other_hints: mo.ui.switch = mo.ui.switch(
        label=utils.ui.small(strings.ui_show_other_hints), value=False
    )
    dlt_schema_show_custom_hints: mo.ui.switch = mo.ui.switch(
        label=utils.ui.small(strings.ui_show_custom_hints), value=False
    )
    dlt_clear_result_cache: mo.ui.run_button = mo.ui.run_button(
        label=utils.ui.small(strings.ui_clear_cache)
    )
    dlt_restrict_to_last_1000: mo.ui.switch = mo.ui.switch(
        label=utils.ui.small(strings.ui_limit_to_1000_rows), value=True
    )
    return (
        dlt_clear_result_cache,
        dlt_restrict_to_last_1000,
        dlt_schema_show_child_tables,
        dlt_schema_show_custom_hints,
        dlt_schema_show_dlt_columns,
        dlt_schema_show_dlt_tables,
        dlt_schema_show_other_hints,
        dlt_schema_show_row_counts,
        dlt_schema_show_type_hints,
        dlt_section_browse_data_switch,
        dlt_section_data_quality_switch,
        dlt_section_ibis_browser_switch,
        dlt_section_loads_switch,
        dlt_section_info_switch,
        dlt_section_schema_switch,
        dlt_section_state_switch,
        dlt_section_sync_switch,
        dlt_section_trace_switch,
    )


@app.cell(hide_code=True)
def watch_changes(
    dlt_pipeline_select: mo.ui.multiselect,
    dlt_pipelines_dir: str,
):
    """
    Watch changes in the trace file and trigger reload in the home cell and all following cells on change
    """
    from dlt.pipeline.trace import get_trace_file_path

    # provide pipeline object to the following cells
    dlt_pipeline_name: str = (
        str(dlt_pipeline_select.value[0]) if dlt_pipeline_select.value else None
    )
    dlt_file_watcher = None
    if dlt_pipeline_name:
        dlt_file_watcher = mo.watch.file(get_trace_file_path(dlt_pipelines_dir, dlt_pipeline_name))
    return dlt_pipeline_name, dlt_file_watcher


@app.cell(hide_code=True)
def ui_primary_controls(
    dlt_pipeline: dlt.Pipeline,
    dlt_schema_show_child_tables: mo.ui.switch,
    dlt_schema_show_dlt_tables: mo.ui.switch,
    dlt_schema_show_row_counts: mo.ui.switch,
    dlt_section_browse_data_switch: mo.ui.switch,
    dlt_section_schema_switch: mo.ui.switch,
    dlt_section_trace_switch: mo.ui.switch,
    dlt_config: DashboardConfiguration,
    dlt_schema_select: mo.ui.multiselect,
):
    """
    Helper cell for creating certain controls based on selected sections
    """

    dlt_selected_schema_name = (
        cast(str, dlt_schema_select.value) if dlt_schema_select.value else None
    )

    #
    # Schema controls
    #
    dlt_schema_table_list: mo.ui.table = None
    if dlt_section_schema_switch.value and dlt_pipeline and dlt_selected_schema_name:
        _table_list = utils.schema.create_table_list(
            dlt_config,
            dlt_pipeline,
            dlt_selected_schema_name,
            show_internals=dlt_schema_show_dlt_tables.value,
            show_child_tables=dlt_schema_show_child_tables.value,
        )
        dlt_schema_table_list = utils.ui.dlt_table(
            _table_list,
            selection="multi",
            initial_selection=[0],
        )

    #
    # Browse data controls
    #
    dlt_data_table_list: mo.ui.table = None
    if dlt_section_browse_data_switch.value and dlt_pipeline and dlt_selected_schema_name:
        table_list = utils.schema.create_table_list(
            dlt_config,
            dlt_pipeline,
            dlt_selected_schema_name,
            show_internals=dlt_schema_show_dlt_tables.value,
            show_child_tables=dlt_schema_show_child_tables.value,
            show_row_counts=dlt_schema_show_row_counts.value,
        )
        dlt_data_table_list = utils.ui.dlt_table(
            table_list,
            selection="single",
            initial_selection=[0],
        )

    #
    # Trace steps table
    #
    dlt_trace_steps_table: mo.ui.table = None
    try:
        if dlt_section_trace_switch.value and dlt_pipeline and dlt_pipeline.last_trace:
            dlt_trace_steps_table = utils.ui.dlt_table(
                utils.trace.trace_steps_overview(dlt_config, dlt_pipeline.last_trace),
                selection="multi",
                freeze_column="step",
            )
    except Exception as exc:
        logger.error(f"Error while building trace steps table: {exc}")
    return (
        dlt_data_table_list,
        dlt_schema_table_list,
        dlt_trace_steps_table,
        dlt_selected_schema_name,
    )


@app.cell(hide_code=True)
def utils_cli_args_and_query_vars_config():
    """
    Prepare cli args  as globals for the following cells
    """
    _run_context = dlt.current.run_context()
    mo_query_var_pipeline_name: str = None
    mo_cli_arg_pipelines_dir: str = None
    mo_cli_arg_with_test_identifiers: bool = False
    mo_cli_arg_pipeline: str = None
    mo_query_var_profile: str = None
    mo_cli_arg_profile: str = None
    try:
        mo_query_var_pipeline_name = cast(str, mo.query_params().get("pipeline")) or None
        mo_cli_arg_pipeline = cast(str, mo.cli_args().get("pipeline")) or None
        mo_cli_arg_pipelines_dir = cast(str, mo.cli_args().get("pipelines-dir")) or None
        mo_cli_arg_with_test_identifiers = (
            cast(bool, mo.cli_args().get("with_test_identifiers")) or False
        )
        mo_query_var_profile = (
            cast(str, mo.query_params().get("profile")) or None
            if isinstance(_run_context, ProfilesRunContext)
            else None
        )
        mo_cli_arg_profile = (
            cast(str, mo.cli_args().get("profile")) or None
            if isinstance(_run_context, ProfilesRunContext)
            else None
        )
    except Exception:
        pass

    return (
        mo_cli_arg_pipelines_dir,
        mo_cli_arg_with_test_identifiers,
        mo_query_var_pipeline_name,
        mo_cli_arg_pipeline,
        mo_query_var_profile,
        mo_cli_arg_profile,
    )


if __name__ == "__main__":
    app.run()



"""The module `dlt.destinations.dataset` is deprecated. Use in `dlt.dataset`

Changes were made backwards-compatible:
- `dlt.destinations.dataset.dataset.ReadableDBAPIDataset` -> `dlt.Dataset`
- `dlt.destinations.dataset.relation.ReadableDBAPIRelation` -> `dlt.Relation`
- `dlt.destinations.dataset::dataset()` -> `dlt.dataset()`
"""

from typing import TYPE_CHECKING

# NOTE this warning is only trigger when type-checking.
# Move it outside the type checking block when closer to deprecation date.
if not TYPE_CHECKING:
    from dlt.common.warnings import DltDeprecationWarning

    DltDeprecationWarning(
        "Content from this module was moved to `dlt._dataset`, which is internal.",
        since="1.16.0",
        expected_due="2.0.0",
    )

from dlt.dataset.dataset import dataset
from dlt.destinations.dataset.dataset import ReadableDBAPIDataset
from dlt.destinations.dataset.relation import ReadableDBAPIRelation

from dlt.destinations.dataset.utils import (
    get_destination_clients,
    get_destination_client_initial_config,
)


__all__ = (
    "dataset",
    "ReadableDBAPIDataset",
    "ReadableDBAPIRelation",
    "get_destination_client_initial_config",
    "get_destination_clients",
)



from typing import Tuple
from dlt.common.exceptions import MissingDependencyException
from dlt.common.destination import AnyDestination, Destination
from dlt.common.destination.client import (
    JobClientBase,
    DestinationClientDwhConfiguration,
    DestinationClientStagingConfiguration,
    DestinationClientConfiguration,
    DestinationClientDwhWithStagingConfiguration,
)
from dlt.common.schema import Schema


def get_destination_client_initial_config(
    destination: AnyDestination,
    dataset_name: str,
    multi_dataset_default_schema_name: str = None,
    as_staging: bool = False,
) -> DestinationClientConfiguration:
    client_spec = destination.spec

    # this client supports many schemas and datasets
    if issubclass(client_spec, DestinationClientDwhConfiguration):
        if issubclass(client_spec, DestinationClientStagingConfiguration):
            config: DestinationClientDwhConfiguration = client_spec(
                as_staging_destination=as_staging
            )
        else:
            config = client_spec()
        config._bind_dataset_name(
            dataset_name, default_schema_name=multi_dataset_default_schema_name
        )
    else:
        config = client_spec()  # type: ignore[assignment]

    return config


def get_destination_clients(
    schema: Schema,
    destination: AnyDestination = None,
    destination_dataset_name: str = None,
    multi_dataset_default_schema_name: str = None,
    staging: AnyDestination = None,
    staging_dataset_name: str = None,
) -> Tuple[JobClientBase, JobClientBase]:
    """Creates destination and/or staging job clients and binds them to `schema` and dataset names.
    Configuration in destination SPEC will be resolved.

    if `multi_dataset_default_schema_name` is set, each `schema` gets own dataset name
    as `dataset_name__{schema.name}`, except the default schema that gets `dataset_name`

    """

    destination = Destination.from_reference(destination) if destination else None
    staging = Destination.from_reference(staging) if staging else None

    try:
        # resolve staging config in order to pass it to destination client config
        staging_client = None
        if staging:
            staging_initial_config = get_destination_client_initial_config(
                staging,
                dataset_name=staging_dataset_name,
                multi_dataset_default_schema_name=multi_dataset_default_schema_name,
                as_staging=True,
            )
            # create the client - that will also resolve the config
            staging_client = staging.client(schema, staging_initial_config)

        initial_config = get_destination_client_initial_config(
            destination,
            dataset_name=destination_dataset_name,
            multi_dataset_default_schema_name=multi_dataset_default_schema_name,
        )

        # attach the staging client config to destination client config - if its type supports it
        if (
            staging_client
            and isinstance(initial_config, DestinationClientDwhWithStagingConfiguration)
            and isinstance(staging_client.config, DestinationClientStagingConfiguration)
        ):
            initial_config.staging_config = staging_client.config
        # create instance with initial_config properly set
        client = destination.client(schema, initial_config)
        return client, staging_client
    except ModuleNotFoundError:
        # TODO: support destinations defined in other packages ie. by importing
        # top level module of destination type and getting PKG_NAME from version module of it
        from dlt import version

        client_spec = destination.spec()
        raise MissingDependencyException(
            f"{client_spec.destination_type} destination",
            [f"{version.DLT_PKG_NAME}[{client_spec.destination_type}]"],
            "Dependencies for specific destinations are available as extras of dlt",
        )



from typing import Optional

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.runners.configuration import PoolRunnerConfiguration, TPoolType
from dlt.common.storages import (
    LoadStorageConfiguration,
    NormalizeStorageConfiguration,
    SchemaStorageConfiguration,
)


@configspec
class ItemsNormalizerConfiguration(BaseConfiguration):
    add_dlt_id: bool = False
    """When true, items to be normalized will have `_dlt_id` column added with a unique ID for each row."""
    add_dlt_load_id: bool = False
    """When true, items to be normalized will have `_dlt_load_id` column added with the current load ID."""


@configspec
class NormalizeConfiguration(PoolRunnerConfiguration):
    pool_type: TPoolType = "process"
    destination_capabilities: DestinationCapabilitiesContext = None  # injectable
    _schema_storage_config: SchemaStorageConfiguration = None
    _normalize_storage_config: NormalizeStorageConfiguration = None
    _load_storage_config: LoadStorageConfiguration = None

    json_normalizer: ItemsNormalizerConfiguration = ItemsNormalizerConfiguration(
        add_dlt_id=True, add_dlt_load_id=True
    )

    parquet_normalizer: ItemsNormalizerConfiguration = ItemsNormalizerConfiguration(
        add_dlt_id=False, add_dlt_load_id=False
    )

    model_normalizer: ItemsNormalizerConfiguration = ItemsNormalizerConfiguration(
        add_dlt_id=False, add_dlt_load_id=True
    )

    def on_resolved(self) -> None:
        self.pool_type = "none" if self.workers == 1 else "process"



from os import environ
from os.path import isdir
from typing import Any, Optional, Type, Tuple

from dlt.common.configuration.specs.base_configuration import is_secret_hint

from .provider import ConfigProvider, get_key_name

SECRET_STORAGE_PATH: str = "/run/secrets/%s"


class EnvironProvider(ConfigProvider):
    @staticmethod
    def get_key_name(key: str, *sections: str) -> str:
        return get_key_name(key, "__", *sections).upper()

    @property
    def name(self) -> str:
        return "Environment Variables"

    def get_value(
        self, key: str, hint: Type[Any], pipeline_name: str, *sections: str
    ) -> Tuple[Optional[Any], str]:
        # apply section to the key
        key = self.get_key_name(key, pipeline_name, *sections)
        if is_secret_hint(hint):
            # try secret storage
            try:
                # must conform to RFC1123 DNS LABELS (https://kubernetes.io/docs/concepts/overview/working-with-objects/names/#dns-label-names)
                secret_name = key.lower().replace("_", "-")
                secret_path = SECRET_STORAGE_PATH % secret_name
                # kubernetes stores secrets as files in a dir, docker compose plainly
                if isdir(secret_path):
                    secret_path += "/" + secret_name
                with open(secret_path, "r", encoding="utf-8") as f:
                    secret = f.read()
                # add secret to environ so forks have access
                # warning: removing new lines is not always good. for password OK for PEMs not
                # warning: in regular secrets that is dealt with in particular configuration logic
                environ[key] = secret.strip()
                # do not strip returned secret
                return secret, key
            # includes FileNotFound
            except OSError:
                pass
        return environ.get(key, None), key

    @property
    def supports_secrets(self) -> bool:
        return True

    @property
    def supports_sections(self) -> bool:
        return True

    @property
    def is_empty(self) -> bool:
        return len(environ) == 0



import contextlib
from typing import Dict, Iterable, List, Optional, Set

from dlt.common import logger
from dlt.common.configuration.inject import with_config
from dlt.common.destination.exceptions import (
    DestinationCapabilitiesException,
    IdentifierTooLongException,
)
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.destination.exceptions import UnsupportedDataType
from dlt.common.destination.capabilities import DestinationCapabilitiesContext, LOADER_FILE_FORMATS
from dlt.common.schema import Schema
from dlt.common.schema.exceptions import (
    SchemaIdentifierNormalizationCollision,
    TableNotFound,
)
from dlt.common.schema.typing import (
    TColumnType,
    TLoaderMergeStrategy,
    TLoaderReplaceStrategy,
    TSchemaTables,
    TTableSchema,
)
from dlt.common.schema.utils import (
    fill_hints_from_parent_and_clone_table,
    get_merge_strategy,
    is_complete_column,
    find_incomplete_columns,
)
from dlt.common.storages import ParsedLoadJobFileName
from dlt.common.typing import ConfigValue, DictStrStr, TLoaderFileFormat


def verify_schema_capabilities(
    schema: Schema,
    capabilities: DestinationCapabilitiesContext,
    destination_type: str,
    warnings: bool = True,
) -> List[Exception]:
    """Verifies `load_tables` that have all hints filled by job client before loading against capabilities.
    Returns a list of exceptions representing critical problems with the schema.
    It will log warnings by default. It is up to the caller to eventually raise exception

    * Checks all table and column name lengths against destination capabilities and raises on too long identifiers
    * Checks if schema has collisions due to case sensitivity of the identifiers
    """

    # collect all exceptions to show all problems in the schema
    exception_log: List[Exception] = []
    # combined casing function
    case_identifier = lambda ident: capabilities.casefold_identifier(
        (str if capabilities.has_case_sensitive_identifiers else str.casefold)(ident)
    )
    table_name_lookup: DictStrStr = {}
    # name collision explanation
    collision_msg = "Destination is case " + (
        "sensitive" if capabilities.has_case_sensitive_identifiers else "insensitive"
    )
    if capabilities.casefold_identifier is not str:
        collision_msg += (
            f" but it uses {capabilities.casefold_identifier} to generate case insensitive"
            " identifiers. You may try to change the destination capabilities by changing the"
            " `casefold_identifier` to `str`"
        )
    collision_msg += (
        ". Please clean up your data before loading so the entities have different name. You can"
        " also change to case insensitive naming convention. Note that in that case data from both"
        " columns will be merged into one."
    )

    # check for any table clashes
    for table in schema.data_tables():
        table_name = table["name"]
        # detect table name conflict
        cased_table_name = case_identifier(table_name)
        if cased_table_name in table_name_lookup:
            conflict_table_name = table_name_lookup[cased_table_name]
            exception_log.append(
                SchemaIdentifierNormalizationCollision(
                    schema.name,
                    table_name,
                    "table",
                    table_name,
                    conflict_table_name,
                    schema.naming.name(),
                    collision_msg,
                )
            )
        table_name_lookup[cased_table_name] = table_name
        if len(table_name) > capabilities.max_identifier_length:
            exception_log.append(
                IdentifierTooLongException(
                    destination_type,
                    "table",
                    table_name,
                    capabilities.max_identifier_length,
                )
            )

        column_name_lookup: DictStrStr = {}
        for column_name in dict(table["columns"]):
            # detect table name conflict
            cased_column_name = case_identifier(column_name)
            if cased_column_name in column_name_lookup:
                conflict_column_name = column_name_lookup[cased_column_name]
                exception_log.append(
                    SchemaIdentifierNormalizationCollision(
                        schema.name,
                        table_name,
                        "column",
                        column_name,
                        conflict_column_name,
                        schema.naming.name(),
                        collision_msg,
                    )
                )
            column_name_lookup[cased_column_name] = column_name
            if len(column_name) > capabilities.max_column_identifier_length:
                exception_log.append(
                    IdentifierTooLongException(
                        destination_type,
                        "column",
                        f"{table_name}.{column_name}",
                        capabilities.max_column_identifier_length,
                    )
                )
    return exception_log


def column_type_to_str(column: TColumnType) -> str:
    """Converts column type to db-like type string"""
    data_type: str = column["data_type"]
    precision = column.get("precision")
    scale = column.get("scale")
    if precision is not None and scale is not None:
        data_type += f"({precision},{scale})"
    elif precision is not None:
        data_type += f"({precision})"
    return data_type


def verify_supported_data_types(
    prepared_tables: Iterable[PreparedTableSchema],
    new_jobs: Iterable[ParsedLoadJobFileName],
    capabilities: DestinationCapabilitiesContext,
    destination_type: str,
    warnings: bool = True,
) -> List[Exception]:
    exception_log: List[Exception] = []
    # can't check types without type mapper
    if capabilities.type_mapper is None or not new_jobs:
        return exception_log

    type_mapper = capabilities.get_type_mapper()

    # index available file formats
    table_file_formats: Dict[str, Set[TLoaderFileFormat]] = {}
    for parsed_file in new_jobs:
        formats = table_file_formats.setdefault(parsed_file.table_name, set())
        if parsed_file.file_format in LOADER_FILE_FORMATS:
            formats.add(parsed_file.file_format)  # type: ignore[arg-type]
    # all file formats
    all_file_formats = set(capabilities.supported_loader_file_formats or []) | set(
        capabilities.supported_staging_file_formats or []
    )

    for table in prepared_tables:
        # map types
        for column in table["columns"].values():
            # do not verify incomplete columns, those won't be created
            if not is_complete_column(column):
                continue
            try:
                type_mapper.to_destination_type(column, table)
            except Exception as ex:
                # collect mapping exceptions
                exception_log.append(ex)
            # ensure if types can be loaded from file formats present in jobs
            for format_ in table_file_formats.get(table["name"], []):
                try:
                    type_mapper.ensure_supported_type(column, table, format_)
                except ValueError as err:
                    # figure out where data type is supported
                    available_in_formats: List[TLoaderFileFormat] = []
                    for candidate_format in all_file_formats - set([format_]):
                        with contextlib.suppress(Exception):
                            type_mapper.ensure_supported_type(column, table, candidate_format)
                            available_in_formats.append(candidate_format)
                    exception_log.append(
                        UnsupportedDataType(
                            destination_type,
                            table["name"],
                            column["name"],
                            column_type_to_str(column),
                            format_,
                            available_in_formats,
                            err.args[0],
                        )
                    )

    return exception_log


@with_config
def prepare_load_table(
    tables: TSchemaTables,
    table: TTableSchema,
    destination_capabilities: DestinationCapabilitiesContext = ConfigValue,
) -> PreparedTableSchema:
    """Prepares a table schema to be loaded by filling missing hints and doing other modifications required by given destination.

    `destination_capabilities` are injected from context if not explicitly passed.

    Returns: prepared table, note: `table` is cloned
    """
    table_name = table["name"]
    try:
        prep_table = fill_hints_from_parent_and_clone_table(tables, table)
        table_format = prep_table.get("table_format")
        # set table format if preferred
        if destination_capabilities.preferred_table_format and not table_format:
            prep_table["table_format"] = destination_capabilities.preferred_table_format
        # if table_format == "preferred":
        #     # NOTE: can also set None as table format
        #     prep_table["table_format"] = destination_capabilities.preferred_table_format
        if table_format == "native":
            # no special table format, use whatever is supported
            prep_table["table_format"] = None
        # remove incomplete columns
        for column, _ in find_incomplete_columns(table):
            prep_table["columns"].pop(column["name"])
        return prep_table  # type: ignore[return-value]
    except KeyError:
        raise TableNotFound("<>", table_name)


def resolve_replace_strategy(
    table: PreparedTableSchema,
    required_strategy: Optional[TLoaderReplaceStrategy],
    destination_capabilities: Optional[DestinationCapabilitiesContext],
) -> Optional[TLoaderReplaceStrategy]:
    """Returns replace strategy for the supplied prepared table and destination capabilities.
    If `required_strategy` cannot be fulfilled or capabilities do not support replace, None is returned.
    """
    supported_replace_strategies = (
        destination_capabilities.replace_strategies_selector(
            destination_capabilities.supported_replace_strategies, table_schema=table
        )
        if destination_capabilities.replace_strategies_selector
        else destination_capabilities.supported_replace_strategies
    )
    if not supported_replace_strategies:
        return None

    if required_strategy and required_strategy not in supported_replace_strategies:
        return None

    return required_strategy or supported_replace_strategies[0]


@with_config
def resolve_merge_strategy(
    tables: TSchemaTables,
    table: TTableSchema,
    destination_capabilities: Optional[DestinationCapabilitiesContext] = ConfigValue,
) -> Optional[TLoaderMergeStrategy]:
    """Resolve merge strategy for a table, possibly resolving the 'x-merge-strategy from a table chain. strategies selector in `destination_capabilities`
    is used if present. If `table` does not contain strategy hint, a default value will be used which is the first.

    `destination_capabilities` are injected from context if not explicitly passed.

    Returns None if table write disposition is not merge
    """
    if table.get("write_disposition") == "merge":
        destination_capabilities = (
            destination_capabilities or DestinationCapabilitiesContext.generic_capabilities()
        )
        supported_strategies = destination_capabilities.supported_merge_strategies
        # prepare table
        table = prepare_load_table(tables, table, destination_capabilities)

        table_name = table["name"]
        if destination_capabilities.merge_strategies_selector:
            supported_strategies = destination_capabilities.merge_strategies_selector(
                supported_strategies, table_schema=table
            )
        if not supported_strategies:
            return None
        merge_strategy = get_merge_strategy(tables, table_name)
        # use first merge strategy as default
        if merge_strategy is None and supported_strategies:
            merge_strategy = supported_strategies[0]
        if merge_strategy not in supported_strategies:
            raise DestinationCapabilitiesException(
                f"`{merge_strategy}` merge strategy not supported for table `{table_name}`. "
                f"Available strategies: {supported_strategies}"
            )
        return merge_strategy
    return None



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Any, Generator, Mapping, Optional, Union

import dlt
from dlt.common.destination.dataset import SupportsDataAccess
from dlt.helpers.ibis import DATA_TYPE_MAP, create_ibis_backend, _get_ibis_to_sqlglot_compiler
from dlt.common.schema import Schema as DltSchema
from dlt.common.destination import TDestinationReferenceArg
from dlt.common.schema.typing import TDataType, TTableSchema
from dlt.common.exceptions import MissingDependencyException

if TYPE_CHECKING:
    import pandas as pd
    import pyarrow as pa
    from ibis import BaseBackend  # noqa: I251

try:
    # ibis imports follow the convention used in the ibis source code
    import ibis  # noqa: I251
    from ibis import util as ibis_util, options as ibis_options  # noqa: I251
    from ibis.expr.types import Expr, Table  # noqa: I251
    import ibis.expr.datatypes as dt
    import ibis.expr.operations as ops
    import ibis.expr.schema as sch
    import ibis.expr.types as ir
    import sqlglot as sg
    import ibis.backends.sql.compilers as sc
    from ibis.backends import NoUrl, NoExampleLoader
    from ibis.backends.sql import SQLBackend
    from ibis.formats import TypeMapper
    from sqlglot import expressions as sge
except ImportError:
    raise MissingDependencyException("dlt ibis helpers", ["ibis-framework"])

# TODO move `dlt.helpers.ibis` content to this module


# TODO complete implementation and register to singledispatch `sch.infer.register`
class DltType(TypeMapper):
    @classmethod
    def from_ibis(cls) -> TDataType:
        raise NotImplementedError

    @classmethod
    def to_ibis(cls, typ: TDataType, nullable: Optional[bool] = None) -> dt.DataType:
        nullable = True if nullable is None else bool(nullable)
        return ibis.dtype(DATA_TYPE_MAP[typ], nullable=nullable)


# def _transpile(query: sge.ExpOrStr, *, target_dialect: type[sg.Dialect]) -> str:
#     if isinstance(query, sg.Expression):
#         query = query.sql(dialect=target_dialect)
#     elif isinstance(query, str):
#         query = sg.transpile(query, write=target_dialect)[0]
#     else:
#         raise TypeErrorWithKnownTypes(
#             key="query", value_received=query, valid_types=["str", "sqlglot.Expression"]
#         )

#     return query


# TODO support `database` kwarg (equiv. `dataset_name`) to enable DltBackend to access multiple database
# NOTE most of the code below was derived from the Ibis BaseBackend and SQLBackend implementation
class _DltBackend(SQLBackend, NoUrl, NoExampleLoader):
    """Ibis backend that delegates execution to dlt's native SQL engine.

    To make Ibis expressions executable, they need to be bound to a table on the
    backend (i.e., data that exists somewhere). By default, Ibis backends work by
    creating and maintaining a connection to the backend.

    This `DltBackend` is "lazier" and doesn't make a connection to the backend where
    the data lives ("backend" in Ibis, "destination" in dlt). Instead, it uses dlt
    metadata to register what table exists and create bound expressions.

    For example, if you use Ibis with the Snowflake backend, you will get use Ibis'
    implementation. If you use the DltBackend with a Snowflake destination, you will
    use the dlt SQL client for Snowflake. The SQL query created should be the same
    because dlt internally uses the `ibis -> sqlglot` compiler, but the actual
    execution might differ. This is especially likely when calling `.execute()` or
    other methods that return data in memory.
    """

    name: str = "dlt"
    supports_temporary_tables = False
    supports_python_udfs = False

    @classmethod
    def from_dataset(cls, dataset: dlt.Dataset) -> _DltBackend:
        """Create an Ibis `DltBackend` from a `dlt.Dataset`

        This enables `dlt.Relation.to_ibis()` to create bound tables.
        """
        # NOTE it's essential to keep this lazy, no round-trip to destination
        # sync with destination should be made through the dataset.
        new_backend = cls()
        new_backend._dataset = dataset
        new_backend.compiler = _get_ibis_to_sqlglot_compiler(dataset.destination_dialect)
        return new_backend

    def to_native_ibis(self, *, read_only: bool = False) -> BaseBackend:
        return get_native_ibis_backend(self._dataset, read_only=read_only)

    @property
    # @override
    def version(self) -> str:
        import importlib.metadata

        return importlib.metadata.version("dlt")

    # NOTE can change signature
    # @override
    def do_connect(
        self,
        destination: TDestinationReferenceArg,
        dataset_name: str,
        schema: Union[DltSchema, str, None] = None,
        **config: Any,
    ) -> None:
        self._dataset = dlt.dataset(
            destination=destination,
            dataset_name=dataset_name,
            schema=schema,
        )
        self.compiler = _get_ibis_to_sqlglot_compiler(self._dataset.destination_dialect)

    def disconnect(self) -> None:
        # no need to disconnect, no connections are persisted
        pass

    @contextlib.contextmanager
    def _safe_raw_sql(
        self, query: Union[sge.ExpOrStr, str, ir.Expr], **kwargs: Any
    ) -> Generator[SupportsDataAccess, None, None]:
        # just re-yield our cursor which is dbapi compatible
        with self._dataset.query(query)._cursor() as cur:
            yield cur

    def compile(  # noqa
        self,
        expr: ir.Expr,
        /,
        *,
        limit: Union[int, str, None] = None,
        params: Optional[Mapping[ir.Scalar, Any]] = None,
        pretty: bool = False,
    ) -> str:
        # this reuses dlt.Relation to generate destination-specific SQL from destination agnostic
        # expr
        r_ = self._dataset.query(expr)
        if limit:
            if limit == "default":
                limit = ibis_options.sql.default_limit
            if limit:
                r_ = r_.limit(int(limit))
        sql = r_.to_sql(pretty=pretty)
        self._log(sql)
        # TODO: allow for `params`
        return sql

    # required for marimo DataSources UI to work
    @property
    def current_database(self) -> str:
        return self._dataset.dataset_name

    # required for marimo DataSources UI to work
    def list_tables(
        self, *, like: Optional[str] = None, database: Union[tuple[str, str], str, None] = None
    ) -> list[str]:
        """Return the list of table names"""
        return list(self._dataset.schema.tables.keys())

    # required for marimo DataSources UI to work
    def get_schema(self, table_name: str, *args: Any, **kwargs: Any) -> sch.Schema:
        return _to_ibis_schema(self._dataset.table(table_name).schema)

    # required for marimo DataSources UI to work
    def _get_schema_using_query(self, query: str) -> sch.Schema:
        return _to_ibis_schema(self._dataset(query).schema)

    # required for marimo DataSources UI to work
    def table(
        self, name: str, /, *, database: Union[tuple[str, str], str, None] = None
    ) -> ir.Table:
        table_schema = self.get_schema(name)
        return ops.DatabaseTable(
            name,
            schema=table_schema,
            source=self,
            namespace=ops.Namespace(database=self.current_database),
        ).to_expr()

    # TODO use the new dlt `model` format with INSERT statement
    # for non-SQL (e.g., filesystem) use JobClient
    def create_table(
        self,
        name: str,
        /,
        obj: Union[pd.DataFrame, pa.Table, ir.Table, None] = None,
        *,
        schema: Optional[sch.Schema] = None,
        database: Optional[str] = None,
        temp: bool = False,
        overwrite: bool = False,
    ) -> ir.Table:
        raise NotImplementedError

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        """Required to subclass SQLBackend"""
        raise NotImplementedError

    def _register_udfs(self, expr: ir.Expr) -> None:
        """Override SQLBackend method to avoid round-trip to Ibis SQL compiler"""

    def _fetch_from_cursor(self, cursor: SupportsDataAccess, schema: sch.Schema) -> pd.DataFrame:
        # SqlBackend implementation is clearly wrong - it passes cursor to `from_records`
        # which expects data
        # as a bonus - this will provide native pandas reading for destinations that support it

        from ibis.formats.pandas import PandasData

        df = PandasData.convert_table(cursor.df(), schema)
        return df

    @ibis_util.experimental
    def to_pyarrow(
        self,
        expr: ir.Expr,
        /,
        *,
        params: Optional[Mapping[ir.Scalar, Any]] = None,
        limit: Union[int, str, None] = None,
        **kwargs: Any,
    ) -> pa.Table:
        self._run_pre_execute_hooks(expr)

        table_expr = expr.as_table()
        schema = table_expr.schema()
        arrow_schema = schema.to_pyarrow()

        with self._safe_raw_sql(self.compile(expr, limit=limit, params=params)) as cur:
            table = cur.arrow()

        return expr.__pyarrow_result__(
            table.rename_columns(list(table_expr.columns)).cast(arrow_schema)
        )

    # TODO: implement native arrow batches
    # @util.experimental
    # def to_pyarrow_batches(


def _to_ibis_schema(table_schema: TTableSchema) -> sch.Schema:
    return sch.Schema(
        {
            name: DltType.to_ibis(column["data_type"], column.get("nullable"))
            for name, column in table_schema["columns"].items()
        }
    )


def get_native_ibis_backend(dataset: dlt.Dataset, *, read_only: bool = False) -> BaseBackend:
    return create_ibis_backend(
        destination=dataset._destination,
        client=dataset.destination_client,
        read_only=read_only,
    )



from dlt.common.destination.capabilities import (
    DestinationCapabilitiesContext,
    merge_caps_file_formats,
    TLoaderFileFormat,
    LOADER_FILE_FORMATS,
)
from dlt.common.destination.reference import (
    TDestinationReferenceArg,
    Destination,
    AnyDestination,
    DestinationReference,
)
from dlt.common.destination.typing import PreparedTableSchema

__all__ = [
    "DestinationCapabilitiesContext",
    "merge_caps_file_formats",
    "TLoaderFileFormat",
    "LOADER_FILE_FORMATS",
    "PreparedTableSchema",
    "TDestinationReferenceArg",
    "Destination",
    "AnyDestination",
    "DestinationReference",
]



import argparse

from dlt.common.configuration.specs.pluggable_run_context import (
    RunContextBase,
)

from dlt._workspace.cli import echo as fmt, utils
from dlt._workspace._workspace_context import WorkspaceRunContext
from dlt._workspace.typing import TLocationInfo
from dlt._workspace.cli.utils import (
    check_delete_local_data,
    delete_local_data,
    fetch_workspace_info,
)
from dlt._workspace.cli._pipeline_command import list_pipelines


def _format_location_tag(loc: TLocationInfo) -> str:
    """Build a human-readable tag for a location (e.g. '(global, profile: dev)')."""
    parts = []
    if loc["scope"] == "global":
        parts.append("global")
    if profile_name := loc.get("profile_name"):
        parts.append("profile: %s" % profile_name)
    if not loc["present"]:
        parts.append("not found")
    if parts:
        return " (%s)" % ", ".join(parts)
    return ""


@utils.track_command("workspace", track_before=False, operation="info")
def print_workspace_info(run_context: WorkspaceRunContext, verbosity: int = 0) -> None:
    info = fetch_workspace_info()

    if info["name"]:
        fmt.echo("Workspace %s:" % fmt.bold(info["name"]))
    fmt.echo("Workspace dir: %s" % fmt.bold(info["run_dir"]))
    fmt.echo("Settings dir: %s" % fmt.bold(info["settings_dir"]))

    profile = info["profile"]
    if profile:
        fmt.echo()
        fmt.echo("Settings for profile %s:" % fmt.bold(profile["name"]))
        fmt.echo("  Pipelines and other working data: %s" % fmt.bold(profile["data_dir"]))
        fmt.echo("  Locally loaded data: %s" % fmt.bold(profile["local_dir"]))
        if profile["is_pinned"]:
            fmt.echo("  Profile is %s" % fmt.bold("pinned"))
        if info["configured_profiles"]:
            fmt.echo(
                "Profiles with configs or pipelines: %s"
                % fmt.bold(", ".join(info["configured_profiles"]))
            )

    fmt.echo()
    fmt.echo("dlt found configuration in following locations:")
    total_not_found_count = 0
    for prov in info["providers"]:
        fmt.echo("* %s" % fmt.bold(prov["name"]))
        for loc in prov["locations"]:
            if loc["present"]:
                tag = _format_location_tag(loc)
                fmt.echo("    %s%s" % (loc["path"], tag))
            else:
                if verbosity > 0:
                    tag = _format_location_tag(loc)
                    fmt.echo("    %s" % fmt.style("%s%s" % (loc["path"], tag), fg="yellow"))
                else:
                    total_not_found_count += 1
        if prov["is_empty"]:
            fmt.echo("    provider is empty")
    if verbosity == 0 and total_not_found_count > 0:
        fmt.echo(
            "%s location(s) were probed but not found. Use %s to see details."
            % (fmt.bold(str(total_not_found_count)), fmt.bold("-v"))
        )
    # installed toolkits
    if info["installed_toolkits"]:
        fmt.echo()
        fmt.echo(
            "Installed toolkits: %s"
            % ", ".join(fmt.bold(n) for n in sorted(info["installed_toolkits"]))
        )

    # list pipelines in the workspace
    fmt.echo()
    list_pipelines(run_context.get_data_entity("pipelines"), verbosity)


@utils.track_command("workspace", track_before=False, operation="clean")
def clean_workspace(run_context: RunContextBase, args: argparse.Namespace) -> None:
    fmt.echo("Local pipelines data will be removed. Remote destinations are not affected.")
    deleted_dirs = check_delete_local_data(run_context, args.skip_data_dir)
    if deleted_dirs:
        delete_local_data(run_context, deleted_dirs)


@utils.track_command("workspace", track_before=True, operation="mcp")
def start_mcp(run_context: WorkspaceRunContext, port: int, stdio: bool, sse: bool) -> None:
    from dlt._workspace.mcp import WorkspaceMCP

    if stdio:
        transport = "stdio"
    elif sse:
        transport = "sse"
    else:
        transport = "streamable-http"
    if transport != "stdio":
        fmt.echo("Starting dlt MCP server", err=True)
    mcp_server = WorkspaceMCP(f"dlt: {run_context.name}@{run_context.profile}", port=port)
    mcp_server.run(transport=transport)


@utils.track_command("dashboard", True)
def show_workspace(run_context: WorkspaceRunContext, edit: bool) -> None:
    from dlt._workspace.helpers.dashboard.runner import run_dashboard

    run_dashboard(edit=edit)



import dataclasses
from typing import ClassVar, List, Final, Optional, Union

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import GcpServiceAccountCredentials, GcpOAuthCredentials
from dlt.common.utils import digest128

from dlt.common.destination.client import DestinationClientDwhWithStagingConfiguration


@configspec
class BigQueryClientConfiguration(DestinationClientDwhWithStagingConfiguration):
    destination_type: Final[str] = dataclasses.field(default="bigquery", init=False, repr=False, compare=False)  # type: ignore
    credentials: Union[GcpServiceAccountCredentials, GcpOAuthCredentials] = None
    location: str = "US"
    project_id: Optional[str] = None
    """Note, that this is BigQuery project_id which could be different from credentials.project_id"""
    has_case_sensitive_identifiers: bool = True
    """If True then dlt expects to load data into case sensitive dataset"""
    should_set_case_sensitivity_on_new_dataset: bool = False
    """If True, dlt will set case sensitivity flag on created datasets that corresponds to naming convention"""

    http_timeout: float = 15.0
    """connection timeout for http request to BigQuery api"""
    file_upload_timeout: float = 30 * 60.0
    """a timeout for file upload when loading local files"""
    retry_deadline: float = 60.0
    """How long to retry the operation in case of error, the backoff 60 s."""
    batch_size: int = 500
    """Number of rows in streaming insert batch"""
    autodetect_schema: bool = False
    """Allow BigQuery to autodetect schemas and create data tables"""
    ignore_unknown_values: bool = False
    """Ignore unknown values in the data"""

    __config_gen_annotations__: ClassVar[List[str]] = ["location"]

    def get_location(self) -> str:
        return self.location

    def fingerprint(self) -> str:
        """Returns a fingerprint of project_id"""
        if self.credentials and self.credentials.project_id:
            return digest128(self.credentials.project_id)
        return ""



from typing import Dict, List, cast

from dlt.common.data_types import TDataType
from dlt.common.typing import DictStrAny
from dlt.common.schema.typing import (
    LOADS_TABLE_NAME,
    VERSION_TABLE_NAME,
    TSimpleRegex,
    TStoredSchema,
    TTableSchemaColumns,
    TColumnDefaultHint,
)
from dlt.common.schema.exceptions import SchemaEngineNoUpgradePathException
from dlt.common.schema.utils import (
    get_columns_names_with_prop,
    new_table,
    version_table,
    loads_table,
    migrate_complex_types,
)


def _remove_seen_null_first(schema_dict: DictStrAny) -> None:
    """Remove legacy seen-null-first columns from all tables.

    Incomplete columns (no data_type) with seen-null-first are removed entirely.
    Complete columns with seen-null-first get the hint stripped.
    """
    for table in schema_dict.get("tables", {}).values():
        for col_name, col_schema in list(table.get("columns", {}).items()):
            x_norm = col_schema.get("x-normalizer", {})
            if x_norm.get("seen-null-first"):
                if "data_type" in col_schema:
                    x_norm.pop("seen-null-first", None)
                    if not x_norm:
                        col_schema.pop("x-normalizer", None)
                else:
                    table["columns"].pop(col_name)


def migrate_schema(schema_dict: DictStrAny, from_engine: int, to_engine: int) -> TStoredSchema:
    if from_engine != to_engine:
        if from_engine == 1 and to_engine > 1:
            schema_dict["includes"] = []
            schema_dict["excludes"] = []
            from_engine = 2
        if from_engine == 2 and to_engine > 2:
            from dlt.common.schema.normalizers import import_normalizers, configured_normalizers

            # current version of the schema
            current = cast(TStoredSchema, schema_dict)
            # add default normalizers and root hash propagation
            # use explicit None to get default settings. ignore any naming conventions
            normalizers = configured_normalizers(naming=None, json_normalizer=None)
            current["normalizers"], _, _ = import_normalizers(normalizers, normalizers)
            current["normalizers"]["json"]["config"] = {
                "propagation": {"root": {"_dlt_id": "_dlt_root_id"}}
            }
            # move settings, convert strings to simple regexes
            d_h: Dict[TColumnDefaultHint, List[TSimpleRegex]] = schema_dict.pop("hints", {})
            for h_k, h_l in d_h.items():
                d_h[h_k] = list(map(lambda r: TSimpleRegex("re:" + r), h_l))
            p_t: Dict[TSimpleRegex, TDataType] = schema_dict.pop("preferred_types", {})
            p_t = {TSimpleRegex("re:" + k): v for k, v in p_t.items()}

            current["settings"] = {
                "default_hints": d_h,
                "preferred_types": p_t,
            }
            # repackage tables
            old_tables: Dict[str, TTableSchemaColumns] = schema_dict.pop("tables")
            current["tables"] = {}
            for name, columns in old_tables.items():
                # find last path separator
                parent = name
                # go back in a loop to find existing parent
                while True:
                    idx = parent.rfind("__")
                    if idx > 0:
                        parent = parent[:idx]
                        if parent not in old_tables:
                            continue
                    else:
                        parent = None
                    break
                nt = new_table(name, parent)
                nt["columns"] = columns
                current["tables"][name] = nt
            # assign exclude and include to tables

            def migrate_filters(group: str, filters: List[str]) -> None:
                # existing filter were always defined at the root table. find this table and move filters
                for f in filters:
                    # skip initial ^
                    root = f[1 : f.find("__")]
                    path = f[f.find("__") + 2 :]
                    t = current["tables"].get(root)
                    if t is None:
                        # must add new table to hold filters
                        t = new_table(root)
                        current["tables"][root] = t
                    t.setdefault("filters", {}).setdefault(group, []).append("re:^" + path)  # type: ignore

            excludes = schema_dict.pop("excludes", [])
            migrate_filters("excludes", excludes)
            includes = schema_dict.pop("includes", [])
            migrate_filters("includes", includes)

            # upgraded
            from_engine = 3
        if from_engine == 3 and to_engine > 3:
            # set empty version hash to pass validation, in engine 4 this hash is mandatory
            schema_dict.setdefault("version_hash", "")
            from_engine = 4
        if from_engine == 4 and to_engine > 4:
            # replace schema versions table
            schema_dict["tables"][VERSION_TABLE_NAME] = version_table()
            schema_dict["tables"][LOADS_TABLE_NAME] = loads_table()
            from_engine = 5
        if from_engine == 5 and to_engine > 5:
            # replace loads table
            schema_dict["tables"][LOADS_TABLE_NAME] = loads_table()
            from_engine = 6
        if from_engine == 6 and to_engine > 6:
            # migrate from sealed properties to schema evolution settings
            schema_dict["settings"].pop("schema_sealed", None)
            for table in schema_dict["tables"].values():
                table.pop("table_sealed", None)
            from_engine = 7
        if from_engine == 7 and to_engine > 7:
            schema_dict["previous_hashes"] = []
            from_engine = 8
        if from_engine == 8 and to_engine > 8:
            # add "seen-data" to all tables with _dlt_id, this will handle packages
            # that are being loaded
            for table in schema_dict["tables"].values():
                if "_dlt_id" in table["columns"]:
                    x_normalizer = table.setdefault("x-normalizer", {})
                    x_normalizer["seen-data"] = True
            from_engine = 9
        if from_engine == 9 and to_engine > 9:
            from dlt.common.schema.normalizers import import_normalizers

            # current = cast(TStoredSchema, schema_dict)

            normalizers = schema_dict["normalizers"]
            _, naming, _ = import_normalizers(normalizers)
            c_dlt_id = naming.normalize_identifier("_dlt_id")
            c_dlt_parent_id = naming.normalize_identifier("_dlt_parent_id")

            for table in schema_dict["tables"].values():
                # migrate complex -> json
                migrate_complex_types(table)
                # modify hints
                if dlt_id_col := table["columns"].get(c_dlt_id):
                    # add row key only if unique is set
                    dlt_id_col["row_key"] = dlt_id_col.get("unique", False)
                if parent_dlt_id_col := table["columns"].get(c_dlt_parent_id):
                    # add parent key
                    parent_dlt_id_col["parent_key"] = parent_dlt_id_col.get("foreign_key", False)
                # drop all foreign keys
                for column in table["columns"].values():
                    column.pop("foreign_key", None)

            # migrate preferred types
            if settings := schema_dict.get("settings"):
                if p_t := settings.get("preferred_types"):
                    for re_ in list(p_t.keys()):
                        if p_t[re_] == "complex":
                            p_t[re_] = "json"
            # migrate default hints
            if default_hints := schema_dict["settings"].get("default_hints"):
                # drop foreign key
                default_hints.pop("foreign_key", None)
                # add row and parent key
                default_hints["row_key"] = [TSimpleRegex(c_dlt_id)]
                default_hints["parent_key"] = [TSimpleRegex(c_dlt_parent_id)]

            # remove `generate_dlt_id` from normalizer
            if json_norm := normalizers.get("json"):
                if json_config := json_norm.get("config"):
                    json_config.pop("generate_dlt_id", None)

            from_engine = 10
        if from_engine == 10 and to_engine > 10:
            schema_dict["normalizers"]["use_break_path_on_normalize"] = False
            from_engine = 11

        schema_dict["engine_version"] = from_engine
        if from_engine != to_engine:
            raise SchemaEngineNoUpgradePathException(
                schema_dict["name"], schema_dict["engine_version"], from_engine, to_engine
            )

    # remove null-first on updated schema
    _remove_seen_null_first(schema_dict)

    return cast(TStoredSchema, schema_dict)



import re
import base64
import hashlib
import warnings
import yaml
from argparse import Namespace
from copy import deepcopy, copy
from typing import Dict, List, Sequence, Tuple, Type, Any, cast, Iterable, Optional, Union

from dlt.common.pendulum import pendulum
from dlt.common.time import ensure_pendulum_datetime_utc
from dlt.common import logger
from dlt.common.json import json
from dlt.common.data_types import TDataType
from dlt.common.exceptions import DictValidationException, ValueErrorWithKnownValues
from dlt.common.normalizers.naming import NamingConvention
from dlt.common.normalizers.naming.snake_case import NamingConvention as SnakeCase
from dlt.common.typing import DictStrAny, REPattern
from dlt.common.validation import TCustomValidator, validate_dict_ignoring_xkeys
from dlt.common.schema import detections
from dlt.common.schema.typing import (
    C_DLT_ID,
    C_DLT_LOAD_ID,
    DLT_NAME_PREFIX,
    C_DLT_LOADS_TABLE_LOAD_ID,
    C_CHILD_PARENT_REF_LABEL,
    C_DESCENDANT_ROOT_REF_LABEL,
    C_ROOT_LOAD_REF_LABEL,
    C_VERSION_SCHEMA_NAME_LABEL,
    C_VERSION_SCHEMA_VERSION_LABEL,
    SCHEMA_ENGINE_VERSION,
    LOADS_TABLE_NAME,
    SIMPLE_REGEX_PREFIX,
    VERSION_TABLE_NAME,
    PIPELINE_STATE_TABLE_NAME,
    ColumnPropInfos,
    TColumnName,
    TFileFormat,
    TPartialTableSchema,
    TSchemaTables,
    TSchemaUpdate,
    TSimpleRegex,
    TStoredSchema,
    TTableProcessingHints,
    TColumnProcessingHints,
    TTableReferenceParam,
    TTableSchema,
    TColumnSchemaBase,
    TColumnSchema,
    TColumnProp,
    TTableFormat,
    TColumnDefaultHint,
    TTableSchemaColumns,
    TTypeDetectionFunc,
    TTypeDetections,
    TWriteDisposition,
    TLoaderMergeStrategy,
    TSchemaContract,
    TSortOrder,
    TTableReference,
)
from dlt.common.schema.exceptions import (
    CannotCoerceColumnException,
    ParentTableNotFoundException,
    TablePropertiesConflictException,
    InvalidSchemaName,
)
from dlt.common.warnings import Dlt100DeprecationWarning


RE_NON_ALPHANUMERIC_UNDERSCORE = re.compile(r"[^a-zA-Z\d_]")
DEFAULT_WRITE_DISPOSITION: TWriteDisposition = "append"


def is_valid_schema_name(name: str) -> bool:
    """Schema name must be a valid python identifier and have max len of 64"""
    return (
        name is not None
        and name.isidentifier()
        and len(name) <= InvalidSchemaName.MAXIMUM_SCHEMA_NAME_LENGTH
    )


def is_nested_table(table: TTableSchema) -> bool:
    """Checks if table is a dlt nested table: connected to parent table via row_key - parent_key reference"""
    # "parent" table hint indicates NESTED table.
    return bool(table.get("parent"))


def may_be_nested(table: TTableSchema) -> bool:
    """Table may be nested if it does not define any primary/merge keys"""
    pks = get_columns_names_with_prop(table, "primary_key", include_incomplete=True)
    mks = get_columns_names_with_prop(table, "merge_key", include_incomplete=True)
    return not pks and not mks


def normalize_schema_name(name: str) -> str:
    """Normalizes schema name by using snake case naming convention. The maximum length is 64 characters"""
    snake_case = SnakeCase(InvalidSchemaName.MAXIMUM_SCHEMA_NAME_LENGTH)
    return snake_case.normalize_identifier(name)


def apply_defaults(stored_schema: TStoredSchema) -> TStoredSchema:
    """Applies default hint values to `stored_schema` in place

    Updates only complete column hints, incomplete columns are preserved intact
    """
    for table_name, table in stored_schema["tables"].items():
        # overwrite name
        table["name"] = table_name
        for column_name in table["columns"]:
            # add default hints to tables
            column = table["columns"][column_name]
            # overwrite column name
            column["name"] = column_name
            # set column with default
            # table["columns"][column_name] = column
        # add default write disposition to root tables
        if not is_nested_table(table):
            if table.get("write_disposition") is None:
                table["write_disposition"] = DEFAULT_WRITE_DISPOSITION
            if table.get("resource") is None:
                table["resource"] = table_name
    return stored_schema


def remove_defaults(stored_schema: TStoredSchema) -> TStoredSchema:
    """Removes default values from `stored_schema` in place, returns the input for chaining

    * removes column and table names from the value
    * removed resource name if same as table name
    """
    clean_tables = deepcopy(stored_schema["tables"])
    for table in clean_tables.values():
        del table["name"]
        # if t.get("resource") == table_name:
        #     del t["resource"]
        for c in table["columns"].values():
            # remove defaults only on complete columns
            # if is_complete_column(c):
            #     remove_column_defaults(c)
            #     # restore "nullable" True because we want to have explicit nullability in stored schema
            #     c["nullable"] = c.get("nullable", True)
            # do not save names
            del c["name"]

    stored_schema["tables"] = clean_tables
    return stored_schema


def has_default_column_prop_value(prop: str, value: Any) -> bool:
    """Checks if `value` is a default for `prop`."""
    # remove all boolean hints that are False, except "nullable" which is removed when it is True
    if prop in ColumnPropInfos:
        return value in ColumnPropInfos[prop].defaults
    # for any unknown hint ie. "x-" the defaults are
    return value in (None, False)


def is_compound_prop(prop: str) -> bool:
    """Checks if a column property is compound."""
    if prop in ColumnPropInfos:
        return ColumnPropInfos[prop].compound is True
    # for any unknown property
    return False


def remove_compound_props(
    columns: TTableSchemaColumns, compound_props: set[str]
) -> TTableSchemaColumns:
    """Removes compound properties from all columns in place.

    Args:
        columns: Table columns to modify.
        compound_props: Set of property names to remove.

    Returns:
        The modified columns dict (same object that was passed in).

    Note: This is a generic property remover, but the name reflects its intended use.
    It removes properties even if their value is False.
    """
    for column in columns.values():
        for prop in compound_props:
            column.pop(prop, None)  # type: ignore[misc]
    return columns


def remove_column_defaults(column_schema: TColumnSchema) -> TColumnSchema:
    """Removes default values from `column_schema` in place, returns the input for chaining"""
    # remove hints with default values
    for h in list(column_schema.keys()):
        if has_default_column_prop_value(h, column_schema[h]):  # type: ignore
            del column_schema[h]  # type: ignore
    return column_schema


def bump_version_if_modified(stored_schema: TStoredSchema) -> Tuple[int, str, str, List[str]]:
    """Bumps the `stored_schema` version and version hash if content modified, returns (new version, new hash, old hash, 10 last hashes) tuple"""
    hash_ = generate_version_hash(stored_schema)
    previous_hash = stored_schema.get("version_hash")
    previous_version = stored_schema.get("version")
    if not previous_hash:
        # if hash was not set, set it without bumping the version, that's initial schema
        # previous_version may not be None for migrating schemas
        stored_schema["version"] = previous_version or 1
    elif hash_ != previous_hash:
        stored_schema["version"] += 1
        store_prev_hash(stored_schema, previous_hash)

    stored_schema["version_hash"] = hash_
    return stored_schema["version"], hash_, previous_hash, stored_schema["previous_hashes"]


def store_prev_hash(
    stored_schema: TStoredSchema, previous_hash: str, max_history_len: int = 10
) -> None:
    # unshift previous hash to previous_hashes and limit array to 10 entries
    previous_hashes = stored_schema["previous_hashes"]
    if previous_hash not in previous_hashes:
        previous_hashes.insert(0, previous_hash)
        if (sur := len(previous_hashes) - max_history_len) > 0:
            del previous_hashes[-sur:]
        # stored_schema["previous_hashes"] = stored_schema["previous_hashes"][:max_history_len]


def generate_version_hash(stored_schema: TStoredSchema) -> str:
    # generates hash out of stored schema content, excluding the hash itself and version
    schema_copy = copy(stored_schema)
    schema_copy.pop("version")
    schema_copy.pop("version_hash", None)
    schema_copy.pop("imported_version_hash", None)
    schema_copy.pop("previous_hashes", None)
    # ignore order of elements when computing the hash
    content = json.dumpb(schema_copy, sort_keys=True)
    h = hashlib.sha3_256(content)
    # additionally check column order
    table_names = sorted((schema_copy.get("tables") or {}).keys())
    if table_names:
        for tn in table_names:
            t = schema_copy["tables"][tn]
            h.update(tn.encode("utf-8"))
            # add column names to hash in order
            for cn in (t.get("columns") or {}).keys():
                h.update(cn.encode("utf-8"))
    return base64.b64encode(h.digest()).decode("ascii")


def verify_schema_hash(
    loaded_schema_dict: DictStrAny, verifies_if_not_migrated: bool = False
) -> bool:
    # generates content hash and compares with existing
    engine_version: str = loaded_schema_dict.get("engine_version")
    # if upgrade is needed, the hash cannot be compared
    if verifies_if_not_migrated and engine_version != SCHEMA_ENGINE_VERSION:
        return True
    # if hash is present we can assume at least v4 engine version so hash is computable
    stored_schema = cast(TStoredSchema, loaded_schema_dict)
    hash_ = generate_version_hash(stored_schema)
    return hash_ == stored_schema["version_hash"]


def normalize_simple_regex_column(naming: NamingConvention, regex: TSimpleRegex) -> TSimpleRegex:
    """Assumes that regex applies to column name and normalizes it."""

    def _normalize(r_: str) -> str:
        is_exact = len(r_) >= 2 and r_[0] == "^" and r_[-1] == "$"
        if is_exact:
            r_ = r_[1:-1]
        # if this a simple string then normalize it
        if r_ == re.escape(r_):
            r_ = naming.normalize_path(r_)
        if is_exact:
            r_ = "^" + r_ + "$"
        return r_

    if regex.startswith(SIMPLE_REGEX_PREFIX):
        return cast(TSimpleRegex, SIMPLE_REGEX_PREFIX + _normalize(regex[3:]))
    else:
        return cast(TSimpleRegex, _normalize(regex))


def canonical_simple_regex(regex: str) -> TSimpleRegex:
    if regex.startswith(SIMPLE_REGEX_PREFIX):
        return cast(TSimpleRegex, regex)
    else:
        return cast(TSimpleRegex, SIMPLE_REGEX_PREFIX + "^" + regex + "$")


def simple_regex_validator(path: str, pk: str, pv: Any, t: Any) -> bool:
    # custom validator on type TSimpleRegex
    if t is TSimpleRegex:
        if not isinstance(pv, str):
            raise DictValidationException(
                f"field `{pk}` value `{pv}` has invalid type `{type(pv).__name__}` while `str` is"
                " expected",
                path,
                t,
                pk,
                pv,
            )
        if pv.startswith(SIMPLE_REGEX_PREFIX):
            # check if regex
            try:
                re.compile(pv[3:])
            except Exception as e:
                raise DictValidationException(
                    f"field `{pk}` value `{pv[3:]}` does not compile as regex: `{str(e)}`",
                    path,
                    t,
                    pk,
                    pv,
                )
        else:
            if RE_NON_ALPHANUMERIC_UNDERSCORE.match(pv):
                raise DictValidationException(
                    f"field `{pk}` value `{pv}` looks like a regex, please prefix with `re:`",
                    path,
                    t,
                    pk,
                    pv,
                )
        # we know how to validate that type
        return True
    else:
        # don't know how to validate this
        return False


def column_name_validator(naming: NamingConvention) -> TCustomValidator:
    def validator(path: str, pk: str, pv: Any, t: Any) -> bool:
        if t is TColumnName:
            if not isinstance(pv, str):
                raise DictValidationException(
                    f"field `{pk}` value `{pv}` has invalid type `{type(pv).__name__}` while"
                    " `str` is expected",
                    path,
                    t,
                    pk,
                    pv,
                )
            try:
                if naming.normalize_path(pv) != pv:
                    raise DictValidationException(
                        f"field `{pk}` value `{pv}` is not a valid column name", path, t, pk, pv
                    )
            except ValueError:
                raise DictValidationException(
                    f"field `{pk}` value `{pv}` is not a valid column name", path, t, pk, pv
                )
            return True
        else:
            return False

    return validator


def _prepare_simple_regex(r: TSimpleRegex) -> str:
    if r.startswith(SIMPLE_REGEX_PREFIX):
        return r[3:]
    else:
        # exact matches
        return "^" + re.escape(r) + "$"


def compile_simple_regex(r: TSimpleRegex) -> REPattern:
    return re.compile(_prepare_simple_regex(r))


def compile_simple_regexes(r: Iterable[TSimpleRegex]) -> REPattern:
    """Compile multiple patterns as one"""
    pattern = "|".join(f"({_prepare_simple_regex(p)})" for p in r)
    if not pattern:  # Don't create an empty pattern that matches everything
        raise ValueError("Cannot create a regex pattern from empty sequence")
    return re.compile(pattern)


def validate_stored_schema(stored_schema: TStoredSchema) -> None:
    # use lambda to verify only non extra fields
    validate_dict_ignoring_xkeys(
        spec=TStoredSchema, doc=stored_schema, path=".", validator_f=simple_regex_validator
    )
    # check child parent relationships
    for table_name, table in stored_schema["tables"].items():
        parent_table_name = table.get("parent")
        if parent_table_name:
            if parent_table_name not in stored_schema["tables"]:
                raise ParentTableNotFoundException(
                    stored_schema["name"], table_name, parent_table_name
                )


def autodetect_sc_type(detection_fs: Sequence[TTypeDetections], t: Type[Any], v: Any) -> TDataType:
    if detection_fs:
        for detection_fn in detection_fs:
            # the method must exist in the module
            detection_f: TTypeDetectionFunc = getattr(detections, "is_" + detection_fn)
            dt = detection_f(t, v)
            if dt is not None:
                return dt
    return None


def is_complete_column(col: TColumnSchemaBase) -> bool:
    """Returns true if column contains enough data to be created at the destination. Must contain a name and a data type. Other hints have defaults."""
    return bool(col.get("name")) and bool(col.get("data_type"))


def is_nullable_column(col: TColumnSchemaBase) -> bool:
    """Returns true if column is nullable"""
    return col.get("nullable", True)


def find_incomplete_columns(table: TTableSchema) -> Iterable[Tuple[TColumnSchemaBase, bool]]:
    """Yields (column, nullable) for all incomplete columns in `table`"""
    for col in table["columns"].values():
        if not is_complete_column(col):
            yield col, is_nullable_column(col)


def compare_complete_columns(a: TColumnSchema, b: TColumnSchema) -> bool:
    """Compares mandatory fields of complete columns"""
    assert is_complete_column(a)
    assert is_complete_column(b)
    return a["data_type"] == b["data_type"] and a["name"] == b["name"]


def compare_table_references(a: TTableReference, b: TTableReference) -> bool:
    if a["referenced_table"] != b["referenced_table"]:
        return False
    a_col_map = dict(zip(a["columns"], a["referenced_columns"]))
    b_col_map = dict(zip(b["columns"], b["referenced_columns"]))
    return a_col_map == b_col_map


def diff_table_references(
    a: Sequence[TTableReference], b: Sequence[TTableReference]
) -> List[TTableReference]:
    """Return a list of references containing references matched by table:
    * References from `b` that are not in `a`
    * References from `b` that are different from the one in `a`
    """
    a_refs_mapping = {ref["referenced_table"]: ref for ref in a}
    new_refs: List[TTableReference] = []
    for b_ref in b:
        table_name = b_ref["referenced_table"]
        if table_name not in a_refs_mapping:
            new_refs.append(b_ref)
        elif not compare_table_references(a_refs_mapping[table_name], b_ref):
            new_refs.append(b_ref)
    return new_refs


def merge_column(
    col_a: TColumnSchema, col_b: TColumnSchema, merge_defaults: bool = True
) -> TColumnSchema:
    """Merges properties from `col_b` into `col_a`, modifying `col_a` in place.

    All properties from `col_b` are copied into `col_a`, potentially overwriting existing values.

    Args:
        col_a: Target column schema that will be modified
        col_b: Source column schema with properties to merge in
        merge_defaults: If False, removes properties with default values from `col_b` before merging.
            This prevents unnecessary default values from being explicitly set in `col_a`.

    Returns:
        The modified col_a (same object that was passed in)
    """
    col_b_clean = col_b if merge_defaults else remove_column_defaults(copy(col_b))
    for n, v in col_b_clean.items():
        col_a[n] = v  # type: ignore

    return col_a


def _collect_and_remove_compound_props(
    source_columns: TTableSchemaColumns,
    target_columns: TTableSchemaColumns,
) -> None:
    """Finds all non-default compound properties in source columns and removes
    those properties from all target columns.

    Args:
        source_columns: Columns to collect compound properties from
        target_columns: Columns to remove compound properties from
    """
    compound_props: set[str] = set()
    for column in source_columns.values():
        compound_props.update(
            prop
            for prop in column
            if is_compound_prop(prop) and not has_default_column_prop_value(prop, column[prop])  # type: ignore[literal-required]
        )
    if compound_props:
        remove_compound_props(columns=target_columns, compound_props=compound_props)


def merge_columns(
    columns_a: TTableSchemaColumns,
    columns_b: TTableSchemaColumns,
    merge_compound_props: bool = True,
) -> TTableSchemaColumns:
    """Merges columns from `columns_b` into `columns_a`, modifying `columns_a` in place.

    For each column in `columns_b`:
    - If column doesn't exist in `columns_a`, it's added
    - If column exists in `columns_a`, properties from `columns_b` are merged into it

    Args:
        columns_a: Target columns dict that will be modified
        columns_b: Source columns dict with columns/properties to merge in
        merge_compound_props: If set to True, compound properties from `columns_b` are merged to `columns_a`,
            If False, compound properties like primary_key and merge_key are replaced entirely
            rather than merged, so `columns_b`'s non-default values fully override `columns_a`'s.

    Returns:
        The modified `columns_a` (same object that was passed in)

    NOTE: Incomplete columns in `columns_a` that become complete in `columns_b` are removed and
    re-added to preserve order.
    """
    if not merge_compound_props:
        _collect_and_remove_compound_props(columns_b, columns_a)

    # remove incomplete columns in table that are complete in diff table
    for col_name, column_b in columns_b.items():
        column_a = columns_a.get(col_name)
        if is_complete_column(column_b):
            if column_a and not is_complete_column(column_a):
                columns_a.pop(col_name)
        if column_a:
            column_b = merge_column(column_a, column_b)
        # set new or updated column
        columns_a[col_name] = column_b
    return columns_a


def ensure_compatible_tables(
    schema_name: str, tab_a: TTableSchema, tab_b: TPartialTableSchema, ensure_columns: bool = True
) -> None:
    """Ensures that `tab_a` and `tab_b` can be merged without conflicts. Conflicts are detected when

    - tables have different names
    - nested tables have different parents
    - tables have any column with incompatible types

    Note: all the identifiers must be already normalized

    """
    if tab_a["name"] != tab_b["name"]:
        raise TablePropertiesConflictException(
            schema_name, tab_a["name"], "name", tab_a["name"], tab_b["name"]
        )
    table_name = tab_a["name"]
    # check if table properties can be merged
    if tab_a.get("parent") != tab_b.get("parent"):
        raise TablePropertiesConflictException(
            schema_name, table_name, "parent", tab_a.get("parent"), tab_b.get("parent")
        )

    if not ensure_columns:
        return

    tab_a_columns = tab_a["columns"]
    for col_b_name, col_b in tab_b["columns"].items():
        if col_b_name in tab_a_columns:
            col_a = tab_a_columns[col_b_name]
            # we do not support changing data types of columns
            if is_complete_column(col_a) and is_complete_column(col_b):
                if not compare_complete_columns(tab_a_columns[col_b_name], col_b):
                    # attempt to update to incompatible columns
                    raise CannotCoerceColumnException(
                        schema_name,
                        table_name,
                        col_b_name,
                        col_b["data_type"],
                        tab_a_columns[col_b_name]["data_type"],
                        None,
                    )


# def compare_tables(tab_a: TTableSchema, tab_b: TTableSchema) -> bool:
#     try:
#         table_name = tab_a["name"]
#         if table_name != tab_b["name"]:
#             raise TablePropertiesConflictException(table_name, "name", table_name, tab_b["name"])
#         diff_table = diff_tables(tab_a, tab_b, ignore_table_name=False)
#         # columns cannot differ
#         return len(diff_table["columns"]) == 0
#     except SchemaException:
#         return False


def merge_table(
    schema_name: str,
    table: TTableSchema,
    partial_table: TPartialTableSchema,
    merge_compound_props: bool = True,
) -> TPartialTableSchema:
    """Merges `partial_table` into `table` in place.

    Merge rules:
    - New columns from `partial_table` are added to `table`
    - Existing columns are updated: properties from `partial_table` overwrite `table`
    - Incomplete columns in `table` that become complete in `partial_table` are removed
      and re-added at the end to preserve column order
    - References are merged by `referenced_table`: new refs are added, existing refs
      for the same table are updated
    - Other table-level properties from `partial_table` overwrite those in `table`
    - Nothing is deleted from `table` (columns, properties) unless explicitly overwritten

    Args:
        schema_name: Name of the schema, used for error messages.
        table: Target table schema that will be modified in place.
        partial_table: Source table schema with columns/properties to merge in.
        merge_compound_props: If True (default), compound properties (primary_key,
            merge_key, partition, cluster) from `partial_table` are merged additively
            with those in `table`. If False, compound properties in `partial_table`
            are authoritative: any compound property present in `table` but absent
            in `partial_table` is removed.

    Returns:
        The `partial_table` argument (unmodified).

    Raises:
        TablePropertiesConflictException: If tables have different names, different
            parents, or if a nested table has a resource property set.
    """
    ensure_compatible_tables(schema_name, table, partial_table, ensure_columns=False)

    # nested tables with the resource property set - this should not really happen
    if is_nested_table(table) and (resource := partial_table.get("resource")):
        raise TablePropertiesConflictException(
            schema_name, table["name"], "resource", resource, table.get("parent")
        )

    updated_columns = merge_columns(
        table["columns"], partial_table["columns"], merge_compound_props
    )

    # merge references: add new/changed refs from partial_table to table
    if "references" in partial_table:
        new_refs = diff_table_references(
            table.get("references", []), partial_table.get("references", [])
        )
        if new_refs:
            existing_refs = {ref["referenced_table"]: ref for ref in table.get("references", [])}
            for ref in new_refs:
                existing_refs[ref["referenced_table"]] = ref
            table["references"] = list(existing_refs.values())

    # apply table-level hints that are new or changed (skip columns, references, unchanged values)
    for k, v in partial_table.items():
        if k in ("columns", "references"):
            continue
        if table.get(k) != v:
            table[k] = v  # type: ignore[literal-required]
    table["columns"] = updated_columns

    return partial_table


def normalize_table_identifiers(table: TTableSchema, naming: NamingConvention) -> TTableSchema:
    """Normalizes all table and column names in `table` schema according to current schema naming convention and returns
    new instance with modified table schema.

    Naming convention like snake_case may produce name collisions with the column names. Colliding column schemas are merged
    where the column that is defined later in the dictionary overrides earlier column.

    Note that resource name is not normalized.
    """

    table = copy(table)
    table["name"] = naming.normalize_tables_path(table["name"])
    parent = table.get("parent")
    if parent:
        table["parent"] = naming.normalize_tables_path(parent)
    columns = table.get("columns")
    if columns:
        new_columns: TTableSchemaColumns = {}
        for c in columns.values():
            c = copy(c)
            origin_c_name = c["name"]
            new_col_name = c["name"] = naming.normalize_path(c["name"])
            # re-index columns as the name changed, if name space was reduced then
            # some columns now collide with each other. so make sure that we merge columns that are already there
            if new_col_name in new_columns:
                new_columns[new_col_name] = merge_column(
                    new_columns[new_col_name], c, merge_defaults=False
                )
                logger.warning(
                    f"In schema {naming} column {origin_c_name} got normalized into"
                    f" {new_col_name} which collides with other column. Both columns got merged"
                    " into one."
                )
            else:
                new_columns[new_col_name] = c
        table["columns"] = new_columns
    references = table.get("references")
    if references:
        new_references = {}
        for ref in references:
            new_ref = copy(ref)
            new_ref["referenced_table"] = naming.normalize_tables_path(ref["referenced_table"])
            new_ref["columns"] = [naming.normalize_path(c) for c in ref["columns"]]
            new_ref["referenced_columns"] = [
                naming.normalize_path(c) for c in ref["referenced_columns"]
            ]
            if new_ref["referenced_table"] in new_references:
                logger.warning(
                    f"In schema {naming} table {table['name']} has multiple references to"
                    f" {new_ref['referenced_table']}. Only the last one is preserved."
                )
            new_references[new_ref["referenced_table"]] = new_ref

        table["references"] = list(new_references.values())
    return table


def has_table_seen_data(table: TTableSchema) -> bool:
    """Checks if normalizer has seen data coming to the table."""
    return "x-normalizer" in table and table["x-normalizer"].get("seen-data", None) is True


def remove_processing_hints(tables: TSchemaTables) -> TSchemaTables:
    """
    Removes processing hints like x-normalizer and x-loader from schema tables and columns.
    Modifies the input tables and returns it for convenience.
    """
    table_hints, col_hints = get_processing_hints(tables)

    # Remove table-level hints
    for table_name, hints in table_hints.items():
        for hint in hints:
            tables[table_name].pop(hint, None)  # type: ignore[misc]

    # Remove column-level hints
    for table_name, cols in col_hints.items():
        for col_name, hints in cols.items():
            for hint in hints:
                tables[table_name]["columns"][col_name].pop(hint, None)  # type: ignore[misc]

    return tables


def has_seen_null_first_hint(column_schema: TColumnSchema) -> bool:
    """Checks if `column_schema` has seen seen-null-first hint set to True in the x-normalizer hints."""
    return bool(column_schema.get("x-normalizer", {}).get("seen-null-first"))


def remove_seen_null_first_hint(column_schema: TColumnSchema) -> TColumnSchema:
    """Removes seen-null-first hint from the x-normalizer hints in `column_schema` in place,
    if the x-normalizer section becomes empty after removing the hint, it is also removed, returns the modified input
    """
    x_normalizer = column_schema.setdefault("x-normalizer", {})
    if x_normalizer.get("seen-null-first"):
        x_normalizer.pop("seen-null-first", None)
    if not x_normalizer:
        column_schema.pop("x-normalizer", None)
    return column_schema


def get_processing_hints(
    tables: TSchemaTables,
) -> Tuple[Dict[str, List[str]], Dict[str, Dict[str, List[str]]]]:
    """
    Finds processing hints from schema tables and columns.

    Returns:
        A tuple containing:
            - A dictionary mapping table names to a list of table-level processing hints (e.g., 'x-normalizer', 'x-loader').
            - A dictionary mapping table names to another dictionary that maps column names to a list of column-level processing hints (e.g., 'x-normalizer').
    """
    table_hints: Dict[str, List[str]] = {}
    col_hints: Dict[str, Dict[str, List[str]]] = {}

    for table in tables.values():  # <- This is the correct line
        table_name = table["name"]

        for hint in TTableProcessingHints.__annotations__.keys():
            if hint in table:
                table_hints.setdefault(table_name, []).append(hint)

        for col_name, col in table["columns"].items():
            for hint in TColumnProcessingHints.__annotations__.keys():
                if hint in col:
                    col_hints.setdefault(table_name, {}).setdefault(col_name, []).append(hint)

    return table_hints, col_hints


def hint_to_column_prop(h: TColumnDefaultHint) -> TColumnProp:
    if h == "not_null":
        return "nullable"
    return h


def get_columns_names_with_prop(
    table: TTableSchema, column_prop: Union[TColumnProp, str], include_incomplete: bool = False
) -> List[str]:
    return [
        c_n
        for c_n, c in table["columns"].items()
        if column_prop in c
        and not has_default_column_prop_value(column_prop, c[column_prop])  # type: ignore[literal-required]
        and (include_incomplete or is_complete_column(c))
    ]


def get_first_column_name_with_prop(
    table: TTableSchema, column_prop: Union[TColumnProp, str], include_incomplete: bool = False
) -> Optional[str]:
    """Returns name of first column in `table` schema with property `column_prop` or None if no such column exists."""
    column_names = get_columns_names_with_prop(table, column_prop, include_incomplete)
    if len(column_names) > 0:
        return column_names[0]
    return None


def has_column_with_prop(
    table: TTableSchema, column_prop: Union[TColumnProp, str], include_incomplete: bool = False
) -> bool:
    """Checks if `table` schema contains column with property `column_prop`."""
    return len(get_columns_names_with_prop(table, column_prop, include_incomplete)) > 0


def get_dedup_sort_tuple(
    table: TTableSchema, include_incomplete: bool = False
) -> Optional[Tuple[str, TSortOrder]]:
    """Returns tuple with dedup sort information.

    First element is the sort column name, second element is the sort order.

    Returns None if "dedup_sort" hint was not provided.
    """
    dedup_sort_col = get_first_column_name_with_prop(table, "dedup_sort", include_incomplete)
    if dedup_sort_col is None:
        return None
    dedup_sort_order = table["columns"][dedup_sort_col]["dedup_sort"]
    return (dedup_sort_col, dedup_sort_order)


def get_validity_column_names(table: TTableSchema) -> List[Optional[str]]:
    return [
        get_first_column_name_with_prop(table, "x-valid-from"),
        get_first_column_name_with_prop(table, "x-valid-to"),
    ]


def get_active_record_timestamp(table: TTableSchema) -> Optional[pendulum.DateTime]:
    # method assumes a column with "x-active-record-timestamp" property exists
    cname = get_first_column_name_with_prop(table, "x-active-record-timestamp")
    hint_val = table["columns"][cname]["x-active-record-timestamp"]  # type: ignore[typeddict-item]
    return None if hint_val is None else ensure_pendulum_datetime_utc(hint_val)


def merge_schema_updates(schema_updates: Sequence[TSchemaUpdate]) -> TSchemaTables:
    aggregated_update: TSchemaTables = {}
    for schema_update in schema_updates:
        for table_name, table_updates in schema_update.items():
            for partial_table in table_updates:
                # aggregate schema updates
                aggregated_table = aggregated_update.setdefault(table_name, partial_table)
                aggregated_table["columns"].update(partial_table["columns"])
    return aggregated_update


def get_inherited_table_hint(
    tables: TSchemaTables, table_name: str, table_hint_name: str, allow_none: bool = False
) -> Any:
    table = tables.get(table_name, {})
    if table_hint_name in table:
        return table[table_hint_name]  # type: ignore[literal-required]

    if is_nested_table(table):
        return get_inherited_table_hint(tables, table.get("parent"), table_hint_name, allow_none)

    if allow_none:
        return None

    raise ValueError(
        f"No table hint `{table_hint_name}` found in the chain of tables for table `{table_name}`."
    )


def get_write_disposition(tables: TSchemaTables, table_name: str) -> TWriteDisposition:
    """Returns table hint of a table if present. If not, looks up into parent table"""
    return cast(
        TWriteDisposition,
        get_inherited_table_hint(tables, table_name, "write_disposition", allow_none=False),
    )


def get_table_format(tables: TSchemaTables, table_name: str) -> TTableFormat:
    return cast(
        TTableFormat, get_inherited_table_hint(tables, table_name, "table_format", allow_none=True)
    )


def get_file_format(tables: TSchemaTables, table_name: str) -> TFileFormat:
    return cast(
        TFileFormat, get_inherited_table_hint(tables, table_name, "file_format", allow_none=True)
    )


def get_merge_strategy(tables: TSchemaTables, table_name: str) -> TLoaderMergeStrategy:
    return cast(
        TLoaderMergeStrategy,
        get_inherited_table_hint(tables, table_name, "x-merge-strategy", allow_none=True),
    )


def fill_hints_from_parent_and_clone_table(
    tables: TSchemaTables, table: TTableSchema
) -> TTableSchema:
    """Takes write disposition and table format from parent tables if not present"""
    # make a copy of the schema so modifications do not affect the original document
    table = deepcopy(table)
    table_name = table["name"]
    if "write_disposition" not in table:
        table["write_disposition"] = get_write_disposition(tables, table_name)
    if "table_format" not in table:
        if table_format := get_table_format(tables, table_name):
            table["table_format"] = table_format
    if "file_format" not in table:
        if file_format := get_file_format(tables, table_name):
            table["file_format"] = file_format
    if "x-merge-strategy" not in table:
        if strategy := get_merge_strategy(tables, table_name):
            table["x-merge-strategy"] = strategy  # type: ignore[typeddict-unknown-key]
    return table


def table_schema_has_type(table: TTableSchema, _typ: TDataType) -> bool:
    """Checks if `table` schema contains column with type _typ"""
    return any(c.get("data_type") == _typ for c in table["columns"].values())


def table_schema_has_type_with_precision(table: TTableSchema, _typ: TDataType) -> bool:
    """Checks if `table` schema contains column with type _typ and precision set"""
    return any(
        c.get("data_type") == _typ and c.get("precision") is not None
        for c in table["columns"].values()
    )


def get_data_and_dlt_tables(tables: TSchemaTables) -> tuple[list[TTableSchema], list[TTableSchema]]:
    """Separate a list of dlt TTableSchema into a two lists: data tables and internal dlt tables.

    This should be equivalent to `dlt.Schema.data_tables()` and `dlt.Schema.dlt_tables()`
    """
    data_tables: list[TTableSchema] = []
    dlt_tables: list[TTableSchema] = []
    for table_name, table in tables.items():
        if table_name in (VERSION_TABLE_NAME, LOADS_TABLE_NAME, PIPELINE_STATE_TABLE_NAME):
            dlt_tables.append(table)
        else:
            data_tables.append(table)

    return data_tables, dlt_tables


def get_dlt_prefix_by_naming_convetion(naming: NamingConvention) -> str:
    """The dlt prefix, used for tables and columns,normalized according to the naming convention"""
    return naming.normalize_table_identifier(DLT_NAME_PREFIX)


def is_dlt_table_or_column(name: str, normalized_dlt_prefix: str) -> bool:
    """
    Check if a table or column name is a dlt internal name by checking if it starts with dlt prefix.

    Args:
        name: The table or column name to check
        normalized_dlt_prefix: The dlt prefix to check against, normalized by the naming convention
    """
    return name.startswith(normalized_dlt_prefix)


def remove_dlt_columns_from_table(
    table_schema: TTableSchema,
    normalized_dlt_prefix: str,
    exclude_dlt_columns: bool = True,
) -> TTableSchema:
    """
    Remove dlt columns from a single table schema.

    Args:
        table_schema: The table schema to filter
        normalized_dlt_prefix: The dlt prefix to filter by, normalized by the naming convention
        exclude_dlt_columns: If True, remove columns whose name starts with the given prefix

    Returns:
        A new table schema with dlt columns optionally filtered out
    """
    # Create a copy of the table schema, preserving all fields except columns
    new_table_schema = cast(TTableSchema, {k: v for k, v in table_schema.items() if k != "columns"})

    if "columns" in table_schema:
        if exclude_dlt_columns:
            new_table_schema["columns"] = {
                col: col_def
                for col, col_def in table_schema["columns"].items()
                if not is_dlt_table_or_column(col, normalized_dlt_prefix)
            }
        else:
            new_table_schema["columns"] = table_schema["columns"]

    return new_table_schema


def exclude_dlt_entities(
    table_schemas: Iterable[TTableSchema],
    normalized_dlt_prefix: str,
    exclude_dlt_tables: bool = True,
    exclude_dlt_columns: bool = True,
) -> List[TTableSchema]:
    """
    Filter out dlt tables and/or dlt columns from a collection of table schemas.

    Args:
        table_schemas: An iterable of table schemas to filter
        normalized_dlt_prefix: The normalized name of the prefix used to denote internal dlt
            columns and tables, according to the used naming convention
        exclude_dlt_tables: If True, remove tables whose name starts with the given prefix
        exclude_dlt_columns: If True, remove columns whose name starts with the given prefix
    Returns:
        List of filtered table schemas.

    Note: dlt supports changing the default prefix, see schema._dlt_tables_prefix attribute to get
        the source of truth for your schema
    """
    filtered_tables: List[TTableSchema] = []

    for table_schema in table_schemas:
        table_name = table_schema["name"]

        # Skip dlt tables if requested
        if exclude_dlt_tables and is_dlt_table_or_column(table_name, normalized_dlt_prefix):
            continue

        # Remove dlt columns if requested
        filtered_table = remove_dlt_columns_from_table(
            table_schema, normalized_dlt_prefix, exclude_dlt_columns
        )
        filtered_tables.append(filtered_table)

    return filtered_tables


def get_root_table(tables: TSchemaTables, table_name: str) -> TTableSchema:
    """Finds root (without parent) of a `table_name` following the nested references (row_key - parent_key)."""
    table = tables[table_name]
    if is_nested_table(table):
        return get_root_table(tables, table.get("parent"))
    return table


def get_nested_tables(
    tables: TSchemaTables,
    table_name: str,
    max_nesting: Optional[int] = None,
    include_self: Optional[bool] = True,
) -> List[TTableSchema]:
    """Get nested tables for table name and return a list of tables ordered by ancestry so the nested tables are always after their parents

    Note that this function follows only NESTED TABLE reference typically expressed on _dlt_parent_id (PARENT_KEY) to _dlt_id (ROW_KEY).

    Args:
        tables (TSchemaTables): A mapping of table names to their table schema definitions. This is used to look up the root table
            and to recursively find its nested child tables by following their "parent" references.
        table_name (str): The name of the root table from which to collect nested tables.
        max_nesting (Optional[int]): If specified, limits the depth of nesting. 0 = only the root table, 1 = root + direct children, etc.
        include_self (Optional[bool]): If False, the root table itself is excluded from the returned list.

    Returns:
        List[TTableSchema]: A list of nested tables.
    """
    if table_name not in tables:
        return []

    chain: List[TTableSchema] = []

    def _child(t: TTableSchema, current_level: int = 0) -> None:
        name = t["name"]

        if include_self or current_level > 0:
            chain.append(t)

        # Stop recursion if we've reached max nesting level
        if max_nesting is not None and current_level >= max_nesting:
            return

        for candidate in tables.values():
            if is_nested_table(candidate) and candidate.get("parent") == name:
                _child(candidate, current_level + 1)

    _child(tables[table_name])
    return chain


def group_tables_by_resource(
    tables: TSchemaTables, pattern: Optional[REPattern] = None
) -> Dict[str, List[TTableSchema]]:
    """Create a dict of resources and their associated tables and descendant tables
    If `pattern` is supplied, the result is filtered to only resource names matching the pattern.
    """
    result: Dict[str, List[TTableSchema]] = {}
    for table in tables.values():
        resource = table.get("resource")
        if resource and (pattern is None or pattern.match(resource)):
            resource_tables = result.setdefault(resource, [])
            resource_tables.extend(get_nested_tables(tables, table["name"]))
    return result


def create_root_child_reference(tables: TSchemaTables, table_name: str) -> TTableReference:
    """Create a Reference between `{table}.{root_key}` and `{root}.{row_key}`"""
    child_table = tables.get(table_name)
    if child_table is None:
        raise ValueErrorWithKnownValues(
            key="table_name", value_received=table_name, valid_values=list(tables.keys())
        )

    if is_nested_table(child_table) is False:
        raise ValueError(f"Table `{table_name}` is already a root table.")

    root_table = get_root_table(tables, table_name)

    child_root_key: str = get_first_column_name_with_prop(child_table, "root_key")
    if child_root_key is None:
        raise ValueError(
            "No `root_key` found for table `{table_name}`. Set `root_key=True` on"
            " the `@dlt.source` producing this data to generate a `_dlt_root_id` column."
        )
    root_row_key: str = get_first_column_name_with_prop(root_table, "row_key")

    return TTableReference(
        label=C_DESCENDANT_ROOT_REF_LABEL,
        cardinality="many_to_one",
        table=table_name,
        columns=[child_root_key],
        referenced_table=root_table["name"],
        referenced_columns=[root_row_key],
    )


def get_all_root_child_references_from_root(
    tables: TSchemaTables, table_name: str
) -> list[TTableReference]:
    root_table = tables.get(table_name)
    if is_nested_table(root_table) is True:
        raise ValueError(f"Table `{table_name}` is not a root table.")

    children_refs = []
    # skip the first table in chain, which is the root; i.e., the current one
    for child_table in get_nested_tables(tables, table_name)[1:]:
        # try/except because a root table may or may not have child with `root_key` enabled
        try:
            child_ref = create_root_child_reference(tables, child_table["name"])
            children_refs.append(child_ref)
        except ValueError:
            pass

    return children_refs


def create_parent_child_reference(tables: TSchemaTables, table_name: str) -> TTableReference:
    """Create a Reference between `{table}.{parent_key}` and `{parent}.{row_key}`"""
    child_table = tables.get(table_name)
    if child_table is None:
        raise ValueErrorWithKnownValues(
            key="table_name", value_received=table_name, valid_values=list(tables.keys())
        )

    parent_table_name = child_table.get("parent")
    if parent_table_name is None:
        raise ValueError(f"Table `{table_name}` is a root table and has no parent.")
    parent_table = tables.get(parent_table_name)

    child_parent_key: str = get_first_column_name_with_prop(child_table, "parent_key")
    parent_row_key: str = get_first_column_name_with_prop(parent_table, "row_key")

    return TTableReference(
        label=C_CHILD_PARENT_REF_LABEL,
        cardinality="many_to_one",
        table=table_name,
        columns=[child_parent_key],
        referenced_table=parent_table_name,
        referenced_columns=[parent_row_key],
    )


def get_all_parent_child_references_from_root(
    tables: TSchemaTables, table_name: str
) -> list[TTableReference]:
    root_table = tables.get(table_name)
    if is_nested_table(root_table) is True:
        raise ValueError(f"Table `{table_name}` is not a root table.")

    children_refs = []
    # skip the first table in chain, which is the root; i.e., the current one
    for child_table in get_nested_tables(tables, table_name)[1:]:
        # try/except because a root table may or may not have child with `root_key` enabled
        try:
            child_ref = create_parent_child_reference(tables, child_table["name"])
            children_refs.append(child_ref)
        except ValueError:
            pass

    return children_refs


def create_load_table_reference(
    table: TTableSchema, *, naming: NamingConvention = None
) -> TTableReference:
    """Create a Reference between `{table}._dlt_load_id` and `_dlt_loads.load_id`"""
    # TODO temporary solution; refactor caller to always explicitly pass `naming`
    naming = naming if naming else Namespace(normalize_identifier=lambda x: x)  # type: ignore[assignment]

    load_id_column = naming.normalize_identifier(C_DLT_LOAD_ID)
    if table["columns"].get(load_id_column) is None:
        raise ValueError(
            f"Table `{table['name']}` is not a root table and has no `{load_id_column}` column."
        )

    return TTableReference(
        label=C_ROOT_LOAD_REF_LABEL,
        cardinality="many_to_one",
        table=table["name"],
        columns=[load_id_column],
        referenced_table=naming.normalize_identifier(LOADS_TABLE_NAME),
        referenced_columns=[naming.normalize_identifier(C_DLT_LOADS_TABLE_LOAD_ID)],
    )


def create_version_and_loads_hash_reference(
    tables: TSchemaTables, *, naming: NamingConvention = None
) -> TTableReference:
    # TODO temporary solution; refactor caller to always explicitly pass `naming`
    naming = naming if naming else Namespace(normalize_identifier=lambda x: x)  # type: ignore[assignment]

    version_table_name = naming.normalize_identifier(VERSION_TABLE_NAME)
    if version_table_name not in tables:
        raise ValueError(
            f"Table `{version_table_name}` not found in tables: `{list(tables.keys())}`"
        )

    load_table_name = naming.normalize_identifier(LOADS_TABLE_NAME)
    if load_table_name not in tables:
        raise ValueError(f"Table `{load_table_name}` not found in tables: `{list(tables.keys())}`")

    return TTableReference(
        label=C_VERSION_SCHEMA_VERSION_LABEL,
        cardinality="one_to_many",
        table=version_table_name,
        columns=[naming.normalize_identifier("version_hash")],
        referenced_table=load_table_name,
        referenced_columns=[naming.normalize_identifier("schema_version_hash")],
    )


def create_version_and_loads_schema_name_reference(
    tables: TSchemaTables, *, naming: NamingConvention = None
) -> TTableReference:
    # TODO temporary solution; refactor caller to always explicitly pass `naming`
    naming = naming if naming else Namespace(normalize_identifier=lambda x: x)  # type: ignore[assignment]

    version_table_name = naming.normalize_identifier(VERSION_TABLE_NAME)
    if version_table_name not in tables:
        raise ValueError(
            f"Table `{version_table_name}` not found in tables: `{list(tables.keys())}`"
        )

    load_table_name = naming.normalize_identifier(LOADS_TABLE_NAME)
    if load_table_name not in tables:
        raise ValueError(f"Table `{load_table_name}` not found in tables: `{list(tables.keys())}`")

    return TTableReference(
        label=C_VERSION_SCHEMA_NAME_LABEL,
        cardinality="many_to_many",
        table=version_table_name,
        columns=[naming.normalize_identifier("schema_name")],
        referenced_table=load_table_name,
        referenced_columns=[naming.normalize_identifier("schema_name")],
    )


def migrate_complex_types(table: TTableSchema, warn: bool = False) -> None:
    if "columns" not in table:
        return
    table_name = table.get("name")
    for col_name, column in table["columns"].items():
        if data_type := column.get("data_type"):
            if data_type == "complex":
                if warn:
                    warnings.warn(
                        f"`complex` data type found on column {col_name} table {table_name} is"
                        " deprecated. Please use `json` type instead.",
                        Dlt100DeprecationWarning,
                        stacklevel=3,
                    )
                column["data_type"] = "json"


def version_table() -> TTableSchema:
    # NOTE: always add new columns at the end of the table so we have identical layout
    # after an update of existing tables (always at the end)
    # set to nullable so we can migrate existing tables
    # WARNING: do not reorder the columns
    table = new_table(
        VERSION_TABLE_NAME,
        columns=[
            {
                "name": "version",
                "data_type": "bigint",
                "nullable": False,
            },
            {"name": "engine_version", "data_type": "bigint", "nullable": False},
            {"name": "inserted_at", "data_type": "timestamp", "nullable": False},
            {"name": "schema_name", "data_type": "text", "nullable": False},
            {"name": "version_hash", "data_type": "text", "nullable": False},
            {"name": "schema", "data_type": "text", "nullable": False},
        ],
    )
    table["write_disposition"] = "skip"
    table["description"] = "Created by DLT. Tracks schema updates"
    return table


def loads_table() -> TTableSchema:
    # NOTE: always add new columns at the end of the table so we have identical layout
    # after an update of existing tables (always at the end)
    # set to nullable so we can migrate existing tables
    # WARNING: do not reorder the columns
    table = new_table(
        LOADS_TABLE_NAME,
        columns=[
            {
                "name": C_DLT_LOADS_TABLE_LOAD_ID,
                "data_type": "text",
                "nullable": False,
                "precision": 64,
            },
            {"name": "schema_name", "data_type": "text", "nullable": True},
            {"name": "status", "data_type": "bigint", "nullable": False},
            {"name": "inserted_at", "data_type": "timestamp", "nullable": False},
            {
                "name": "schema_version_hash",
                "data_type": "text",
                "nullable": True,
            },
        ],
    )
    table["write_disposition"] = "skip"
    table["description"] = "Created by DLT. Tracks completed loads"
    return table


def dlt_id_column() -> TColumnSchema:
    """Definition of dlt id column"""
    return {
        "name": C_DLT_ID,
        "data_type": "text",
        "nullable": False,
        "unique": True,
        "row_key": True,
        "precision": 64,
    }


def dlt_load_id_column() -> TColumnSchema:
    """Definition of dlt load id column"""
    return {"name": "_dlt_load_id", "data_type": "text", "nullable": False, "precision": 64}


def pipeline_state_table(add_dlt_id: bool = False) -> TTableSchema:
    # NOTE: always add new columns at the end of the table so we have identical layout
    # after an update of existing tables (always at the end)
    # set to nullable so we can migrate existing tables
    # WARNING: do not reorder the columns
    columns: List[TColumnSchema] = [
        {"name": "version", "data_type": "bigint", "nullable": False},
        {"name": "engine_version", "data_type": "bigint", "nullable": False},
        {"name": "pipeline_name", "data_type": "text", "nullable": False},
        {"name": "state", "data_type": "text", "nullable": False},
        {"name": "created_at", "data_type": "timestamp", "nullable": False},
        {"name": "version_hash", "data_type": "text", "nullable": True},
        dlt_load_id_column(),
    ]
    if add_dlt_id:
        columns.append(dlt_id_column())
    table = new_table(
        PIPELINE_STATE_TABLE_NAME,
        write_disposition="append",
        columns=columns,
        # always use caps preferred file format for processing
        file_format="preferred",
    )
    table["description"] = "Created by DLT. Tracks pipeline state"
    return table


def new_table(
    table_name: str,
    parent_table_name: str = None,
    write_disposition: TWriteDisposition = None,
    columns: Sequence[TColumnSchema] = None,
    validate_schema: bool = False,
    resource: str = None,
    schema_contract: TSchemaContract = None,
    table_format: TTableFormat = None,
    file_format: TFileFormat = None,
    references: Sequence[TTableReference] = None,
) -> TTableSchema:
    table: TTableSchema = {
        "name": table_name,
        "columns": {} if columns is None else {c["name"]: c for c in columns},
    }

    if write_disposition:
        table["write_disposition"] = write_disposition
    if schema_contract is not None:
        table["schema_contract"] = schema_contract
    if table_format:
        table["table_format"] = table_format
    if file_format:
        table["file_format"] = file_format
    if references:
        table["references"] = references
    if resource:
        table["resource"] = resource
    if parent_table_name:
        table["parent"] = parent_table_name
    else:
        # set only for root tables
        if not write_disposition:
            # set write disposition only for root tables
            table["write_disposition"] = DEFAULT_WRITE_DISPOSITION
        table["resource"] = resource or table_name

    # migrate complex types to json
    migrate_complex_types(table, warn=True)

    if validate_schema:
        validate_dict_ignoring_xkeys(
            spec=TColumnSchema,
            doc=table["columns"],
            path=f"new_table/{table_name}",
        )
    return table


def new_column(
    column_name: str,
    data_type: TDataType = None,
    nullable: bool = True,
    precision: int = None,
    scale: int = None,
    validate_schema: bool = False,
) -> TColumnSchema:
    column: TColumnSchema = {"name": column_name, "nullable": nullable}
    if data_type:
        column["data_type"] = data_type
    if precision is not None:
        column["precision"] = precision
    if scale is not None:
        column["scale"] = scale
    if validate_schema:
        validate_dict_ignoring_xkeys(
            spec=TColumnSchema,
            doc=column,
            path=f"new_column/{column_name}",
        )

    return column


def default_hints() -> Dict[TColumnDefaultHint, List[TSimpleRegex]]:
    return None


def standard_type_detections() -> List[TTypeDetections]:
    return ["iso_timestamp"]


def to_pretty_json(stored_schema: TStoredSchema) -> str:
    return json.dumps(stored_schema, pretty=True)


def to_pretty_yaml(stored_schema: TStoredSchema) -> str:
    return yaml.dump(stored_schema, allow_unicode=True, default_flow_style=False, sort_keys=False)
